# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2023       John Ralls
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
"""
GrampsTranslations encapsulates Gettext behavior, allowing for a primary
language binding for the user interface and user-invoked language bindings
for reports so that they can be created in a different language from the UI's.
"""

# ------------------------------------------------------------------------
#
# python modules
#
# ------------------------------------------------------------------------
from __future__ import annotations
import gettext
import collections
import inspect


# -------------------------------------------------------------------------
#
# Translations Classes
#
# -------------------------------------------------------------------------
class Lexeme(str):
    r"""
    Created with :meth:`~GrampsTranslations.lexgettext`

    .. rubric:: Example

    Python code::

        _ = lexgettext
        dec = _("|December", "localized lexeme inflections")
        xmas = _("|Christmas", "lexeme")
        text = _("{holiday} is celebrated in {month}".format(
                    holiday=xmas, month=dec))
        greeting = _("Merry {holiday}!").format(holiday=xmas)
        XMAS = xmas.upper()
        print ("\n".join([XMAS, text, greeting]))

    Translation database (Russian example)::

        msgctxt "localized lexeme inflections"
        msgid "|December"
        msgstr "NOMINATIVE=декабрь|GENITIVE=декабря|ABLATIVE=декабрём|"
        "LOCATIVE=декабре"

        msgctxt "lexeme"
        msgid "|Christmas"
        msgstr "NOMINATIVE=рождество|GENITIVE=рождества|ABLATIVE=рождеством"

        msgid "{holiday} is celebrated in {month}"
        msgstr "{holiday} празднуют в {month.forms[LOCATIVE]}"

        msgid "Merry {holiday}!"
        msgstr "Счастливого {holiday.forms[GENITIVE]}!"

    Prints out::

        In English locale:
            CHRISTMAS
            Christmas is celebrated in December
            Merry Christmas!

        In Russian locale:
            РОЖДЕСТВО
            рождество празднуют в декабре
            Счастливого рождества!

    .. rubric:: Description

    Stores an arbitrary number of forms, e.g., inflections.
    These forms are accessible under dictionary keys for each form.
    The names of the forms are language-specific. They are assigned
    by the human translator of the corresponding language (in XX.po)
    as in the example above,
    see :meth:`~GrampsTranslations.lexgettext` docs
    for more info.

    The translated format string can then refer to a specific form
    of the lexeme using ``.``:attr:`~Lexeme.forms` and square brackets:
    ``{holiday.forms[GENITIVE]}``
    expects holiday to be a Lexeme which has a form ``'GENITIVE'`` in it.

    An instance of Lexeme can also be used as a regular unicode string.
    In this case, the work will be delegated to the string for the very
    first form provided in the translated string. In the example above,
    ``{holiday}`` in the translated string will expand to the Russian
    nominative form for Christmas, and ``xmas.upper()`` will produce
    the same nominative form in capital letters.

    .. rubric:: Motivation

    Lexeme is the term used in linguistics for the set of forms taken
    by a particular word, e.g. cases for a noun or tenses for a verb.

    Gramps often needs to compose sentences from several blocks of
    text and single words, often by using python string formatting.

    For instance, formatting a date range is done similarly to this::

        _("Between {startdate_month} {startdate_year}"
              "and {enddate_month} {enddate_year}").format(
                 startdate_month = m1,
                 startdate_year = y1,
                 enddate_month = m2,
                 enddate_year = y2)

    To make such text translatable, the arguments injected into
    format string need to bear all the linguistical information
    on how to plug them into a sentence, i.e., the forms, depending
    on the linguistic context of where the argument appears.
    The format string needs to select the relevant linguistic form.
    This is why ``m1`` and ``m2`` are instances of :class:`~Lexeme`.

    On the other hand, for languages where there is no linguistic
    variation in such sentences, the code needs not to be aware of
    the underlying :class:`~Lexeme` complexity;
    and so they can be processed just like simple strings
    both when passed around in the code and when formatted.
    """

    def __new__(cls, iterable, *args, **kwargs):
        if isinstance(iterable, str):
            newobj = str.__new__(cls, iterable, *args, **kwargs)
        else:
            forms = collections.OrderedDict(iterable)
            values = list(forms.values()) or [""]
            newobj = str.__new__(cls, values[0], *args, **kwargs)
            newobj._forms = forms
        return newobj

    def variants(self):
        """All lexeme forms, in the same order as given upon construction.
        The first one returned is the default form, which is used when the
        Lexeme instance is used in lieu of a string object.

        Same as ``f.values()``"""
        return self._forms.values()

    @property
    def forms(self):
        """Dictionary of the lexeme forms"""
        return self._forms


class GrampsTranslations(gettext.GNUTranslations):
    """
    Overrides and extends gettext.GNUTranslations. See the Python gettext
    "Class API" documentation for how to use this.
    """

    CONTEXT = "%s\x04%s"

    def __init__(self, fp=None):
        super().__init__(fp)
        self.lang = None

    def language(self):
        """
        Return the target languge of this translations object.
        """
        return self.lang

    def gettext(self, msgid: str, context: str = "") -> str:
        """
        Obtain translation of gettext, return a unicode object

        :param msgid: The string to translated.
        :type msgid: unicode
        :param context: The message context.
        :type context: unicode
        :returns: Translation or the original.
        :rtype: unicode
        """
        # If context=="" and msgid =="" then gettext will return po file header
        # and that's not what we want.
        if len((context + msgid).strip()) == 0:
            return msgid
        if context:
            return self.pgettext(context, msgid)
        return gettext.GNUTranslations.gettext(self, msgid)

    def ngettext(self, singular, plural, num):
        """
        The translation of singular/plural is returned unless the translation is
        not available and the singular contains the separator. In that case,
        the returned value is the singular.

        :param singular: The singular form of the string to be translated.
                         may contain a context seperator
        :type singular: unicode
        :param plural: The plural form of the string to be translated.
        :type plural: unicode
        :param num: the amount for which to decide the translation
        :type num: int
        :returns: Translation or the original.
        :rtype: unicode
        """
        return gettext.GNUTranslations.ngettext(self, singular, plural, num)

    def sgettext(self, msgid, context=""):
        """
        Strip the context used for resolving translation ambiguities.

        The translation of msgid is returned unless the translation is
        not available and the msgid contains the separator. In that case,
        the returned value is the portion of msgid following the last
        separator. Default separator is '|'.

        :param msgid: The string to translated.
        :type msgid: unicode
        :param context: The message context.
        :type context: unicode
        :param sep: The separator marking the context.
        :type sep: unicode
        :returns: Translation or the original with context stripped.
        :rtype: unicode
        """
        if "\x04" in msgid:  # Deferred translation
            context, msgid = msgid.split("\x04")
        return self.gettext(msgid, context)

    def lexgettext(self, msgid, context=""):
        """
        Extract all inflections of the same lexeme,
        stripping the '|'-separated context using :meth:`~sgettext`

        The *resulting* message provided by the translator
        is supposed to be '|'-separated as well.
        The possible formats are either (1) a single string
        for a language with no inflections, or (2) a list of
        <inflection name>=<inflected form>, separated with '|'.
        For example:

           (1) "Uninflectable"
           (2) "n=Inflected-nominative|g=Inflected-genitive|d=Inflected-dative"

        See :class:`~Lexeme` documentation for detailed explanation and example.

        :param msgid: The string to translated.
        :type msgid: unicode
        :param context: The message context.
        :type context: unicode
        :returns: Translation or the original with context stripped.
        :rtype: unicode (for option (1)) / Lexeme (option (2))
        """
        variants = self.sgettext(msgid, context).split("|")
        return (
            Lexeme([v.split("=") for v in variants])
            if len(variants) > 1
            else variants[0]
        )

    def pgettext(self, context, message):
        """
        Copied from python 3.8
        """
        ctxt_msg_id = self.CONTEXT % (context, message)
        missing = object()
        tmsg = self._catalog.get(ctxt_msg_id, missing)
        if tmsg is missing:
            if self._fallback:
                return self._fallback.pgettext(context, message)
            return message
        return tmsg


class GrampsNullTranslations(gettext.NullTranslations):
    """
    Extends gettext.NullTranslations to provide the sgettext method.

    Note that it's necessary for msgid to be unicode. If it's not,
    neither will be the returned string.
    """

    def gettext(self, msgid: str, context: str = "") -> str:
        """
        Apply the context if there is one, otherwise just pass it on.
        """
        if context:
            return self.pgettext(context, msgid)
        return gettext.NullTranslations.gettext(self, msgid)

    def sgettext(self, msgid, context=""):
        """
        An old workaround to not having proper context msgids: The msgid
        is a single string with an ASCII Message Separator character
        between the context and the mssgid."
        """
        if "\x04" in msgid:  # Deferred translation
            context, msgid = msgid.split("\x04")
        return self.gettext(msgid, context)

    lexgettext = sgettext

    def language(self):
        """
        The null translation returns the raw msgids, which are in English
        """
        return "en"

    def pgettext(self, context, message):
        """
        Copied from python 3.8
        """
        if self._fallback:
            return self._fallback.pgettext(context, message)
        return message



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2013      Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ....lib import Event
from ....db import Database


# -------------------------------------------------------------------------
#
# HasDayOfWeek
#
# -------------------------------------------------------------------------
class HasDayOfWeek(Rule):
    """Rule that matches an event occurring on a particular day of the week."""

    labels = [_("Day of Week:")]
    name = _("Events occurring on a particular day of the week")
    description = _("Matches events occurring on a particular day of the week")
    category = _("General filters")

    def apply_to_one(self, db: Database, event: Event) -> bool:
        if not self.list[0]:
            return False
        else:
            if event.date:
                dow = event.date.get_dow()
                return dow == int(self.list[0])
            return False



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009 Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
from gramps.gen.plug._pluginreg import newplugin, STABLE, RELCALC
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

MODULE_VERSION = "6.0"

#
# Relationship calculators
#

# ca
plg = newplugin()
plg.id = "relcalc_ca"
plg.name = _("Catalan Relationship Calculator")
plg.description = _("Calculates relationships between people")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "rel_ca.py"
plg.ptype = RELCALC
plg.relcalcclass = "RelationshipCalculator"
## TODO PYTHON 3: does this third entry work?
plg.lang_list = ["ca_ES", "ca", "català", "Catalan", "ca_FR", "ca_AD", "ca_IT"]

# cs
plg = newplugin()
plg.id = "relcalc_cs"
plg.name = _("Czech Relationship Calculator")
plg.description = _("Calculates relationships between people")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "rel_cs.py"
plg.ptype = RELCALC
plg.relcalcclass = "RelationshipCalculator"
plg.lang_list = [
    "cs",
    "CZ",
    "cs_CZ",
    "česky",
    "czech",
    "Czech",
    "cs_CZ.UTF8",
    "cs_CZ.UTF-8",
    "cs_CZ.utf-8",
    "cs_CZ.utf8",
]

# da
plg = newplugin()
plg.id = "relcalc_da"
plg.name = _("Danish Relationship Calculator")
plg.description = _("Calculates relationships between people")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "rel_da.py"
plg.ptype = RELCALC
plg.relcalcclass = "RelationshipCalculator"
plg.lang_list = [
    "da",
    "DA",
    "da_DK",
    "danish",
    "Danish",
    "da_DK.UTF8",
    "da_DK@euro",
    "da_DK.UTF8@euro",
    "dansk",
    "Dansk",
    "da_DK.UTF-8",
    "da_DK.utf-8",
    "da_DK.utf8",
    "da_DK.ISO-8859-1",
    "da_DK.iso-8859-1",
    "da_DK.iso88591",
]

# de
plg = newplugin()
plg.id = "relcalc_de"
plg.name = _("German Relationship Calculator")
plg.description = _("Calculates relationships between people")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "rel_de.py"
plg.ptype = RELCALC
plg.relcalcclass = "RelationshipCalculator"
plg.lang_list = [
    "de",
    "DE",
    "de_DE",
    "deutsch",
    "Deutsch",
    "de_DE.UTF8",
    "de_DE@euro",
    "de_DE.UTF8@euro",
    "german",
    "German",
    "de_DE.UTF-8",
    "de_DE.utf-8",
    "de_DE.utf8",
    "de_CH.UTF-8",
    "de_CH.utf-8",
    "de_CH.utf8",
]

# de_at
plg = newplugin()
plg.id = "relcalc_de_at"
plg.name = _("Austrian Relationship Calculator")
plg.description = _("Calculates relationships between people")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "rel_de_AT.py"
plg.ptype = RELCALC
plg.relcalcclass = "RelationshipCalculator"
plg.lang_list = [
    "de",
    "AT",
    "de_AT",
    "deutsch",
    "Deutsch",
    "de_AT.UTF-8",
    "de_AT.utf-8",
    "de_AT.utf8",
    "german",
    "German",
]

# es
plg = newplugin()
plg.id = "relcalc_es"
plg.name = _("Spanish Relationship Calculator")
plg.description = _("Calculates relationships between people")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "rel_es.py"
plg.ptype = RELCALC
plg.relcalcclass = "RelationshipCalculator"
plg.lang_list = [
    "es",
    "ES",
    "es_ES",
    "espanol",
    "Espanol",
    "es_ES.UTF8",
    "es_ES@euro",
    "es_ES.UTF8@euro",
    "spanish",
    "Spanish",
    "es_ES.UTF-8",
    "es_ES.utf-8",
    "es_ES.utf8",
]

# fi
plg = newplugin()
plg.id = "relcalc_fi"
plg.name = _("Finnish Relationship Calculator")
plg.description = _("Calculates relationships between people")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "rel_fi.py"
plg.ptype = RELCALC
plg.relcalcclass = "RelationshipCalculator"
plg.lang_list = [
    "fi",
    "FI",
    "fi_FI",
    "finnish",
    "Finnish",
    "fi_FI.UTF8",
    "fi_FI@euro",
    "fi_FI.UTF8@euro",
    "suomi",
    "Suomi",
    "fi_FI.UTF-8",
    "fi_FI.utf-8",
    "fi_FI.utf8",
]

# fr
plg = newplugin()
plg.id = "relcalc_fr"
plg.name = _("French Relationship Calculator")
plg.description = _("Calculates relationships between people")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "rel_fr.py"
plg.ptype = RELCALC
plg.relcalcclass = "RelationshipCalculator"
plg.lang_list = [
    "fr",
    "FR",
    "fr_FR",
    "fr_CA",
    "français",
    "Français",
    "fr_FR.UTF8",
    "fr_FR@euro",
    "fr_FR.UTF8@euro",
    "french",
    "French",
    "fr_FR.UTF-8",
    "fr_FR.utf-8",
    "fr_FR.utf8",
    "fr_CA.UTF-8",
    "Français_France",
]

# he
plg = newplugin()
plg.id = "relcalc_he"
plg.name = _("Hebrew Relationship Calculator")
plg.description = _("Calculates relationships between people")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "rel_he.py"
plg.ptype = RELCALC
plg.relcalcclass = "RelationshipCalculator"
plg.lang_list = [
    "עברית",
    "Ivrit",
    "hebrew",
    "Hebrew",
    "he",
    "HE",
    "he_IL",
    "he_IL.UTF-8",
    "he_IL.utf-8",
    "Hebrew_Israel",
]

# hr
plg = newplugin()
plg.id = "relcalc_hr"
plg.name = _("Croatian Relationship Calculator")
plg.description = _("Calculates relationships between people")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "rel_hr.py"
plg.ptype = RELCALC
plg.relcalcclass = "RelationshipCalculator"
plg.lang_list = [
    "hrvatski",
    "Hrvatski",
    "croatian",
    "Croatian",
    "hr",
    "HR",
    "hr_HR",
    "hr_HR.UTF-8",
    "hr_HR.utf-8",
    "Croatian_Croatia",
]

# hu
plg = newplugin()
plg.id = "relcalc_hu"
plg.name = _("Hungarian Relationship Calculator")
plg.description = _("Calculates relationships between people")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "rel_hu.py"
plg.ptype = RELCALC
plg.relcalcclass = "RelationshipCalculator"
plg.lang_list = ["hu", "HU", "hu_HU", "hu_HU.utf8", "hu_HU.UTF8"]

# is
plg = newplugin()
plg.id = "relcalc_is"
plg.name = _("Icelandic Relationship Calculator")
plg.description = _("Calculates relationships between people")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "rel_is.py"
plg.ptype = RELCALC
plg.relcalcclass = "RelationshipCalculator"
plg.lang_list = ["is", "IS", "is_IS", "is_IS@euro", "is_IS.utf8"]

# it
plg = newplugin()
plg.id = "relcalc_it"
plg.name = _("Italian Relationship Calculator")
plg.description = _("Calculates relationships between people")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "rel_it.py"
plg.ptype = RELCALC
plg.relcalcclass = "RelationshipCalculator"
plg.lang_list = ["it", "IT", "it_IT", "it_IT@euro", "it_IT.utf8"]

# nl
plg = newplugin()
plg.id = "relcalc_nl"
plg.name = _("Dutch Relationship Calculator")
plg.description = _("Calculates relationships between people")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "rel_nl.py"
plg.ptype = RELCALC
plg.relcalcclass = "RelationshipCalculator"
plg.lang_list = [
    "nl",
    "NL",
    "nl_NL",
    "nl_BE",
    "nederlands",
    "Nederlands",
    "nl_NL.UTF8",
    "nl_BE.UTF8",
    "nl_NL@euro",
    "nl_NL.UTF8@euro",
    "nl_BE@euro",
    "dutch",
    "Dutch",
    "nl_NL.UTF-8",
    "nl_BE.UTF-8",
    "nl_NL.utf-8",
    "nl_BE.utf-8",
    "nl_NL.utf8",
    "nl_BE.UTF-8",
    "nl_BE.UTF8@euro",
]

# no
plg = newplugin()
plg.id = "relcalc_no"
plg.name = _("Norwegian Relationship Calculator")
plg.description = _("Calculates relationships between people")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "rel_no.py"
plg.ptype = RELCALC
plg.relcalcclass = "RelationshipCalculator"
plg.lang_list = [
    "nb",
    "nn",
    "no",
    "nb_NO",
    "nn_NO",
    "no_NO",
    "nb_NO.UTF8",
    "nn_NO.UTF8",
    "no_NO.UTF8",
    "nb_NO.UTF-8",
    "nn_NO.UTF-8",
    "no_NO.UTF-8",
    "nb_NO.utf-8",
    "nn_NO.utf-8",
    "no_NO.utf-8",
    "nb_NO.iso88591",
    "nn_NO.iso88591",
    "no_NO.iso88591",
    "nynorsk",
    "Nynorsk",
]

# pl
plg = newplugin()
plg.id = "relcalc_pl"
plg.name = _("Polish Relationship Calculator")
plg.description = _("Calculates relationships between people")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "rel_pl.py"
plg.ptype = RELCALC
plg.relcalcclass = "RelationshipCalculator"
plg.lang_list = [
    "pl",
    "PL",
    "pl_PL",
    "polski",
    "Polski",
    "pl_PL.UTF-8",
    "pl_PL.UTF8",
    "pl_PL.utf-8",
    "pl_PL.utf8",
    "pl_PL.iso-8859-2",
    "pl_PL.iso8859-2",
    "pl_PL.cp1250",
    "pl_PL.cp-1250",
]

# pt
plg = newplugin()
plg.id = "relcalc_pt"
plg.name = _("Portuguese Relationship Calculator")
plg.description = _("Calculates relationships between people")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "rel_pt.py"
plg.ptype = RELCALC
plg.relcalcclass = "RelationshipCalculator"
plg.lang_list = [
    "pt",
    "PT",
    "pt_PT",
    "pt_BR",
    "portugues",
    "Portugues",
    "pt_PT.UTF8",
    "pt_BR.UTF8",
    "pt_PT@euro",
    "pt_PT.UTF8@euro",
    "pt_PT.UTF-8",
    "pt_BR.UTF-8",
    "pt_PT.utf-8",
    "pt_BR.utf-8",
    "pt_PT.utf8",
    "pt_BR.utf8",
]

# ru
plg = newplugin()
plg.id = "relcalc_ru"
plg.name = _("Russian Relationship Calculator")
plg.description = _("Calculates relationships between people")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "rel_ru.py"
plg.ptype = RELCALC
plg.relcalcclass = "RelationshipCalculator"
plg.lang_list = [
    "ru",
    "RU",
    "ru_RU",
    "koi8r",
    "ru_koi8r",
    "russian",
    "Russian",
    "ru_RU.koi8r",
    "ru_RU.KOI8-R",
    "ru_RU.utf8",
    "ru_RU.UTF8",
    "ru_RU.utf-8",
    "ru_RU.UTF-8",
    "ru_RU.iso88595",
    "ru_RU.iso8859-5",
    "ru_RU.iso-8859-5",
]

# sk
plg = newplugin()
plg.id = "relcalc_sk"
plg.name = _("Slovak Relationship Calculator")
plg.description = _("Calculates relationships between people")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "rel_sk.py"
plg.ptype = RELCALC
plg.relcalcclass = "RelationshipCalculator"
plg.lang_list = [
    "sk",
    "SK",
    "sk_SK",
    "slovensky",
    "slovak",
    "Slovak",
    "sk_SK.UTF8",
    "sk_SK.UTF-8",
    "sk_SK.utf-8",
    "sk_SK.utf8",
]

# sl
plg = newplugin()
plg.id = "relcalc_sl"
plg.name = _("Slovenian Relationship Calculator")
plg.description = _("Calculates relationships between people")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "rel_sl.py"
plg.ptype = RELCALC
plg.relcalcclass = "RelationshipCalculator"
plg.lang_list = [
    "sl",
    "SL",
    "sl_SI",
    "slovenščina",
    "slovenian",
    "Slovenian",
    "sl_SI.UTF8",
    "sl_SI.UTF-8",
    "sl_SI.utf-8",
    "sl_SI.utf8",
]
# sv
plg = newplugin()
plg.id = "relcalc_sv"
plg.name = _("Swedish Relationship Calculator")
plg.description = _("Calculates relationships between people")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "rel_sv.py"
plg.ptype = RELCALC
plg.relcalcclass = "RelationshipCalculator"
plg.lang_list = [
    "sv",
    "SV",
    "sv_SE",
    "swedish",
    "Swedish",
    "sv_SE.UTF8",
    "sv_SE@euro",
    "sv_SE.UTF8@euro",
    "svenska",
    "Svenska",
    "sv_SE.UTF-8",
    "sv_SE.utf-8",
    "sv_SE.utf8",
    "Swedish_Sweden",
]

# tr
plg = newplugin()
plg.id = "relcalc_tr"
plg.name = _("Turkish Relationship Calculator")
plg.description = _("Calculates relationship between people")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "rel_tr.py"
plg.ptype = RELCALC
plg.relcalcclass = "RelationshipCalculator"
plg.lang_list = [
    "tr",
    "TR",
    "tr_TR",
    "türkçe",
    "Türkçe",
    "turkish",
    "Turkish",
    "tr_TR.utf8",
    "tr_TR.UTF8",
    "tr_TR.utf-8",
    "tr_TR.UTF-8",
]

# uk
plg = newplugin()
plg.id = "relcalc_uk"
plg.name = _("Ukrainian Relationship Calculator")
plg.description = _("Calculates relationships between people")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "rel_uk.py"
plg.ptype = RELCALC
plg.relcalcclass = "RelationshipCalculator"
plg.lang_list = [
    "uk",
    "UA",
    "uk_UA",
    "ukrainian",
    "Ukrainian",
    "uk_UA.utf8",
    "uk_UA.UTF8",
    "uk_UA.utf-8",
    "uk_UA.UTF-8",
    "uk_UA.iso88595",
    "uk_UA.iso8859-5",
    "uk_UA.iso-8859-5",
    "koi8u",
    "uk_koi8u",
    "uk_UA.koi8u",
    "uk_UA.KOI8-U",
]



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2017,2024       Nick Hall
# Copyright (C) 2024-2025       Doug Blank
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Serialization utilities for Gramps.
"""

# ------------------------------------------------------------------------
#
# Python modules
#
# ------------------------------------------------------------------------
import pickle
import logging

LOG = logging.getLogger(".serialize")

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
import gramps.gen.lib as lib
from .json_utils import (
    dict_to_string,
    data_to_object,
    string_to_object,
    string_to_data,
    string_to_dict,
    object_to_data,
    object_to_dict,
    object_to_string,
    data_to_string,
)


class BlobSerializer:
    """
    Serializer for blob data

    In this serializer, data is a nested array,
    and string is pickled bytes.
    """

    data_field = "blob_data"
    metadata_field = "value"

    @staticmethod
    def get_from_data_by_name(data, item):
        if item == "handle":
            return data[0]
        else:
            raise ValueError(f"unknown data name: {item}")

    @staticmethod
    def data_to_object(data, obj_class):
        LOG.debug("blob, data_to_object: %s(%r)", obj_class, data)
        return obj_class.create(data)

    @staticmethod
    def string_to_object(obj_class, bytes):
        LOG.debug("blob, string_to_object: %r...", bytes[:35])
        return obj_class.create(pickle.loads(bytes))

    @staticmethod
    def string_to_data(bytes):
        LOG.debug("blob, string_to_data: %r...", bytes[:35])
        return pickle.loads(bytes)

    @staticmethod
    def object_to_string(obj):
        LOG.debug("blob, object_to_string: %r", obj)
        return pickle.dumps(obj.serialize())

    @staticmethod
    def data_to_string(data):
        LOG.debug("blob, data_to_string: %r", data)
        return pickle.dumps(data)

    @staticmethod
    def metadata_to_object(blob):
        return pickle.loads(blob)

    @staticmethod
    def object_to_metadata(value):
        return pickle.dumps(value)


class JSONSerializer:
    """
    Serializer for JSON data.

    In this serializer, data is a dict,
    and string is a JSON string.
    """

    data_field = "json_data"
    metadata_field = "json_data"

    @staticmethod
    def get_from_data_by_name(data, item):
        if item == "handle":
            return data["handle"]
        else:
            raise ValueError(f"unknown data name: {item}")

    @staticmethod
    def data_to_object(data, obj_class=None):
        LOG.debug("json, data_to_object: %r", data)
        return data_to_object(data)

    @staticmethod
    def string_to_object(obj_class, string):
        LOG.debug("json, string_to_object: %r...", string[:65])
        return string_to_object(string)

    @staticmethod
    def string_to_data(string):
        LOG.debug("json, string_to_data: %r...", string[:65])
        return string_to_data(string)

    @staticmethod
    def object_to_string(obj):
        LOG.debug("json, object_to_string: %r", obj)
        return object_to_string(obj)

    @staticmethod
    def object_to_data(obj):
        LOG.debug("json, object_to_string: %r", obj)
        return object_to_data(obj)

    @staticmethod
    def data_to_string(data):
        LOG.debug("json, data_to_string: %r", data)
        return data_to_string(data)

    @staticmethod
    def metadata_to_object(string):
        doc = string_to_dict(string)
        type_name = doc["type"]
        if type_name == "set":
            return set(doc["value"])
        elif type_name == "tuple":
            return tuple(doc["value"])
        elif type_name == "Researcher":
            return data_to_object(doc["value"])
        else:
            return doc["value"]

    @staticmethod
    def object_to_metadata(value):
        type_name = type(value).__name__
        if type_name in ("set", "tuple"):
            value = list(value)
        elif type_name == "Researcher":
            value = object_to_dict(value)

        data = {
            "type": type_name,
            "value": value,
        }
        return dict_to_string(data)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007       Zsolt Foldvari
# Copyright (C) 2008-2009  Brian G. Matherly
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Printing interface based on Gtk.Print* classes."""

# ------------------------------------------------------------------------
#
# Python modules
#
# ------------------------------------------------------------------------
from math import radians
import logging

# -------------------------------------------------------------------------
#
# GTK modules
#
# -------------------------------------------------------------------------
import cairo

try:  # the Gramps-Connect server has no DISPLAY
    from gi.repository import GObject
    from gi.repository import Gtk
except:
    pass

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.plug.docgen import PAPER_PORTRAIT
import gramps.plugins.lib.libcairodoc as libcairodoc
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
LOG = logging.getLogger(".GtkPrint")

# ------------------------------------------------------------------------
#
# Constants
#
# ------------------------------------------------------------------------

# printer settings (might be needed to align for different platforms)
PRINTER_DPI = 72.0
PRINTER_SCALE = 1.0

# the print settings to remember between print sessions
PRINT_SETTINGS = None

# minimum spacing around a page in print preview
MARGIN = 6

# zoom modes in print preview
(
    ZOOM_BEST_FIT,
    ZOOM_FIT_WIDTH,
    ZOOM_FREE,
) = list(range(3))

# ------------------------------------------------------------------------
#
# Converter functions
#
# ------------------------------------------------------------------------


def paperstyle_to_pagesetup(paper_style):
    """Convert a PaperStyle instance into a Gtk.PageSetup instance.

    @param paper_style: Gramps paper style object to convert
    @param type: PaperStyle
    @return: page_setup
    @rtype: Gtk.PageSetup
    """
    # paper size names according to 'PWG Candidate Standard 5101.1-2002'
    # ftp://ftp.pwg.org/pub/pwg/candidates/cs-pwgmsn10-20020226-5101.1.pdf
    gramps_to_gtk = {
        "Letter": "na_letter",
        "Legal": "na_legal",
        "A0": "iso_a0",
        "A1": "iso_a1",
        "A2": "iso_a2",
        "A3": "iso_a3",
        "A4": "iso_a4",
        "A5": "iso_a5",
        "B0": "iso_b0",
        "B1": "iso_b1",
        "B2": "iso_b2",
        "B3": "iso_b3",
        "B4": "iso_b4",
        "B5": "iso_b5",
        "B6": "iso_b6",
        "B": "na_ledger",
        "C": "na_c",
        "D": "na_d",
        "E": "na_e",
    }

    # First set the paper size
    gramps_paper_size = paper_style.get_size()
    gramps_paper_name = gramps_paper_size.get_name()

    # All sizes not included in the translation table (even if a standard size)
    # are handled as custom format, because we are not intelligent enough.
    if gramps_paper_name in gramps_to_gtk:
        paper_size = Gtk.PaperSize.new(name=gramps_to_gtk[gramps_paper_name])
        LOG.debug("Selected paper size: %s", gramps_to_gtk[gramps_paper_name])
    else:
        if paper_style.get_orientation() == PAPER_PORTRAIT:
            paper_width = gramps_paper_size.get_width() * 10
            paper_height = gramps_paper_size.get_height() * 10
        else:
            paper_width = gramps_paper_size.get_height() * 10
            paper_height = gramps_paper_size.get_width() * 10
        paper_size = Gtk.PaperSize.new_custom(
            "custom", "Custom Size", paper_width, paper_height, Gtk.Unit.MM
        )
        LOG.debug("Selected paper size: (%f,%f)", paper_width, paper_height)

    page_setup = Gtk.PageSetup()
    page_setup.set_paper_size(paper_size)

    # Set paper orientation
    if paper_style.get_orientation() == PAPER_PORTRAIT:
        page_setup.set_orientation(Gtk.PageOrientation.PORTRAIT)
    else:
        page_setup.set_orientation(Gtk.PageOrientation.LANDSCAPE)

    # Set paper margins
    page_setup.set_top_margin(paper_style.get_top_margin() * 10, Gtk.Unit.MM)
    page_setup.set_bottom_margin(paper_style.get_bottom_margin() * 10, Gtk.Unit.MM)
    page_setup.set_left_margin(paper_style.get_left_margin() * 10, Gtk.Unit.MM)
    page_setup.set_right_margin(paper_style.get_right_margin() * 10, Gtk.Unit.MM)

    return page_setup


# ------------------------------------------------------------------------
#
# PrintPreview class
#
# ------------------------------------------------------------------------
class PrintPreview:
    """Implement a dialog to show print preview."""

    zoom_factors = {
        0.50: "50%",
        0.75: "75%",
        1.00: "100%",
        1.25: "125%",
        1.50: "150%",
        1.75: "175%",
        2.00: "200%",
        3.00: "300%",
        4.00: "400%",
    }

    def __init__(self, operation, preview, context, parent):
        self._operation = operation
        self._preview = preview
        self._context = context
        self._parent = parent

        self.__build_window()
        self._current_page = None

    # Private

    def __build_window(self):
        """Build the window from Glade."""
        from gramps.gui.glade import Glade

        glade_xml = Glade()
        self._window = glade_xml.toplevel
        self._window.set_transient_for(self._parent)

        # remember active widgets for future use
        self._swin = glade_xml.get_object("swin")
        self._drawing_area = glade_xml.get_object("drawingarea")
        self._first_button = glade_xml.get_object("first")
        self._prev_button = glade_xml.get_object("prev")
        self._next_button = glade_xml.get_object("next")
        self._last_button = glade_xml.get_object("last")
        self._pages_entry = glade_xml.get_object("entry")
        self._pages_label = glade_xml.get_object("label")
        self._zoom_fit_width_button = glade_xml.get_object("zoom_fit_width")
        self._zoom_best_fit_button = glade_xml.get_object("zoom_best_fit")
        self._zoom_in_button = glade_xml.get_object("zoom_in")
        self._zoom_out_button = glade_xml.get_object("zoom_out")

        # connect the signals
        glade_xml.connect_signals(self)
        self._drawing_area.connect("draw", self.on_drawingarea_draw_event)

    ##def create_surface(self):
    ##return cairo.PDFSurface(StringIO(),
    ##self._context.get_width(),
    ##self._context.get_height())

    ##def get_page(self, page_no):
    ##"""Get the cairo surface of the given page.

    ##Surfaces are also cached for instant access.

    ##"""
    ##if page_no >= len(self._page_numbers):
    ##LOG.debug("Page number %d doesn't exist." % page_no)
    ##page_no = 0

    ##if page_no not in self._page_surfaces:
    ##surface = self.create_surface()
    ##cr = cairo.Context(surface)

    ##if PRINTER_SCALE != 1.0:
    ##cr.scale(PRINTER_SCALE, PRINTER_SCALE)

    ##self._context.set_cairo_context(cr, PRINTER_DPI, PRINTER_DPI)
    ##self._preview.render_page(self._page_numbers[page_no])

    ##self._page_surfaces[page_no] = surface

    ##return self._page_surfaces[page_no]

    def __set_page(self, page_no):
        if page_no < 0 or page_no >= self._page_no:
            return

        if self._current_page != page_no:
            self._drawing_area.queue_draw()

        self._current_page = page_no

        self._first_button.set_sensitive(self._current_page)
        self._prev_button.set_sensitive(self._current_page)
        self._next_button.set_sensitive(self._current_page < self._page_no - 1)
        self._last_button.set_sensitive(self._current_page < self._page_no - 1)

        self._pages_entry.set_text("%d" % (self._current_page + 1))

    def __set_zoom(self, zoom):
        self._zoom = zoom

        screen_width = int(self._paper_width * self._zoom + 2 * MARGIN)
        screen_height = int(self._paper_height * self._zoom + 2 * MARGIN)
        self._drawing_area.set_size_request(screen_width, screen_height)
        self._drawing_area.queue_draw()

        self._zoom_in_button.set_sensitive(self._zoom != max(self.zoom_factors))
        self._zoom_out_button.set_sensitive(self._zoom != min(self.zoom_factors))

    def __zoom_in(self):
        zoom = [z for z in self.zoom_factors if z > self._zoom]

        if zoom:
            return min(zoom)
        else:
            return self._zoom

    def __zoom_out(self):
        zoom = [z for z in self.zoom_factors if z < self._zoom]

        if zoom:
            return max(zoom)
        else:
            return self._zoom

    def __zoom_fit_width(self):
        width, height, vsb_w, hsb_h = self.__get_view_size()

        zoom = width / self._paper_width
        if self._paper_height * zoom > height:
            zoom = (width - vsb_w) / self._paper_width

        return zoom

    def __zoom_best_fit(self):
        width, height, vsb_w, hsb_h = self.__get_view_size()

        zoom = min(width / self._paper_width, height / self._paper_height)

        return zoom

    def __get_view_size(self):
        """Get the dimensions of the scrolled window."""
        width = self._swin.get_allocated_width() - 2 * MARGIN
        height = self._swin.get_allocated_height() - 2 * MARGIN

        if self._swin.get_shadow_type() != Gtk.ShadowType.NONE:
            width -= 2 * self._swin.get_style().xthickness
            height -= 2 * self._swin.get_style().ythickness

        spacing = GObject.Value()
        spacing.init(GObject.TYPE_INT)
        spacing = self._swin.style_get_property("scrollbar-spacing", spacing)
        if spacing:
            spacing = spacing.get_int()
        else:
            spacing = 0

        reqmin, req = self._swin.get_vscrollbar().get_preferred_size()
        vsb_w = spacing + req.width
        reqmin, req = self._swin.get_hscrollbar().get_preferred_size()
        hsb_h = spacing + req.height

        return width, height, vsb_w, hsb_h

    def __end_preview(self):
        self._operation.end_preview()

    # Signal handlers

    def on_drawingarea_draw_event(self, drawing_area, context):
        cr = context
        # cr.rectangle(event.area)
        # cr.clip()

        # get the extents of the page and the screen
        paper_w = int(self._paper_width * self._zoom)
        paper_h = int(self._paper_height * self._zoom)

        width, height, vsb_w, hsb_h = self.__get_view_size()
        if paper_h > height:
            width -= vsb_w
        if paper_w > width:
            height -= hsb_h

        # put the paper on the middle of the window
        xtranslate = MARGIN
        if paper_w < width:
            xtranslate += (width - paper_w) / 2

        ytranslate = MARGIN
        if paper_h < height:
            ytranslate += (height - paper_h) / 2

        cr.translate(xtranslate, ytranslate)

        # draw an empty white page
        cr.set_source_rgb(1.0, 1.0, 1.0)
        cr.rectangle(0, 0, paper_w, paper_h)
        cr.fill_preserve()
        cr.set_source_rgb(0, 0, 0)
        cr.set_line_width(1)
        cr.stroke()

        if self._orientation == Gtk.PageOrientation.LANDSCAPE:
            cr.rotate(radians(-90))
            cr.translate(-paper_h, 0)

        ##page_setup = self._context.get_page_setup()
        ##cr.translate(page_setup.get_left_margin(Gtk.Unit.POINTS),
        ##page_setup.get_top_margin(Gtk.Unit.POINTS))

        ##cr.set_source_surface(self.get_page(0))
        ##cr.paint()

        # draw the content of the currently selected page
        #     Here we use dpi scaling instead of scaling the cairo context,
        #     because it gives better result. In the latter case the distance
        #     of glyphs was changing.
        dpi = PRINTER_DPI * self._zoom
        self._context.set_cairo_context(cr, dpi, dpi)
        self._preview.render_page(self._current_page)

    def on_swin_size_allocate(self, scrolledwindow, allocation):
        if self._zoom_mode == ZOOM_FIT_WIDTH:
            self.__set_zoom(self.__zoom_fit_width())

        if self._zoom_mode == ZOOM_BEST_FIT:
            self.__set_zoom(self.__zoom_best_fit())

    def on_print_clicked(self, toolbutton):
        pass

    def on_first_clicked(self, toolbutton):
        self.__set_page(0)

    def on_prev_clicked(self, toolbutton):
        self.__set_page(self._current_page - 1)

    def on_next_clicked(self, toolbutton):
        self.__set_page(self._current_page + 1)

    def on_last_clicked(self, toolbutton):
        self.__set_page(self._page_no - 1)

    def on_entry_activate(self, entry):
        try:
            new_page = int(entry.get_text()) - 1
        except ValueError:
            new_page = self._current_page

        if new_page < 0 or new_page >= self._page_no:
            new_page = self._current_page

        self.__set_page(new_page)

    def on_zoom_fit_width_toggled(self, toggletoolbutton):
        if toggletoolbutton.get_active():
            self._zoom_best_fit_button.set_active(False)
            self._zoom_mode = ZOOM_FIT_WIDTH
            self.__set_zoom(self.__zoom_fit_width())
        else:
            self._zoom_mode = ZOOM_FREE

    def on_zoom_best_fit_toggled(self, toggletoolbutton):
        if toggletoolbutton.get_active():
            self._zoom_fit_width_button.set_active(False)
            self._zoom_mode = ZOOM_BEST_FIT
            self.__set_zoom(self.__zoom_best_fit())
        else:
            self._zoom_mode = ZOOM_FREE

    def on_zoom_in_clicked(self, toolbutton):
        self._zoom_fit_width_button.set_active(False)
        self._zoom_best_fit_button.set_active(False)
        self._zoom_mode = ZOOM_FREE
        self.__set_zoom(self.__zoom_in())

    def on_zoom_out_clicked(self, toolbutton):
        self._zoom_fit_width_button.set_active(False)
        self._zoom_best_fit_button.set_active(False)
        self._zoom_mode = ZOOM_FREE
        self.__set_zoom(self.__zoom_out())

    def on_window_delete_event(self, widget, event):
        self.__end_preview()
        return False

    def on_quit_clicked(self, toolbutton):
        self.__end_preview()
        self._window.destroy()

    # Public

    def start(self):
        # get paper/page dimensions
        page_setup = self._context.get_page_setup()
        self._paper_width = page_setup.get_paper_width(Gtk.Unit.POINTS)
        self._paper_height = page_setup.get_paper_height(Gtk.Unit.POINTS)
        self._page_width = page_setup.get_page_width(Gtk.Unit.POINTS)
        self._page_height = page_setup.get_page_height(Gtk.Unit.POINTS)
        self._orientation = page_setup.get_orientation()

        # get the total number of pages
        ##self._page_numbers = [0,]
        ##self._page_surfaces = {}
        self._page_no = self._operation.get_property("n_pages")
        self._pages_label.set_text(_("of %d") % self._page_no)

        # set zoom level and initial page number
        self._zoom_mode = ZOOM_FREE
        self.__set_zoom(1.0)
        self.__set_page(0)

        # let's the show begin...
        self._window.show()


# ------------------------------------------------------------------------
#
# GtkPrint class
#
# ------------------------------------------------------------------------
class GtkPrint(libcairodoc.CairoDoc):
    """Print document via GtkPrint* interface.

    Requires Gtk+ 2.10.

    """

    def run(self):
        """Run the Gtk Print operation."""
        global PRINT_SETTINGS

        # get a page setup from the paper style we have
        page_setup = paperstyle_to_pagesetup(self.paper)

        # set up a print operation
        operation = Gtk.PrintOperation()
        operation.set_default_page_setup(page_setup)
        operation.connect("begin_print", self.on_begin_print)
        operation.connect("draw_page", self.on_draw_page)
        operation.connect("paginate", self.on_paginate)
        operation.connect("preview", self.on_preview)

        # set print settings if it was stored previously
        if PRINT_SETTINGS is not None:
            operation.set_print_settings(PRINT_SETTINGS)

        # run print dialog
        while True:
            self.preview = None
            res = operation.run(
                Gtk.PrintOperationAction.PRINT_DIALOG, self.uistate.window
            )
            if self.preview is None:  # cancel or print
                break
            # set up printing again; can't reuse PrintOperation?
            operation = Gtk.PrintOperation()
            operation.set_default_page_setup(page_setup)
            operation.connect("begin_print", self.on_begin_print)
            operation.connect("draw_page", self.on_draw_page)
            operation.connect("paginate", self.on_paginate)
            operation.connect("preview", self.on_preview)
            # set print settings if it was stored previously
            if PRINT_SETTINGS is not None:
                operation.set_print_settings(PRINT_SETTINGS)

        # store print settings if printing was successful
        if res == Gtk.PrintOperationResult.APPLY:
            PRINT_SETTINGS = operation.get_print_settings()

    def on_begin_print(self, operation, context):
        """Setup environment for printing."""
        # get data from context here only once to save time on pagination
        self.page_width = round(context.get_width())
        self.page_height = round(context.get_height())
        self.dpi_x = context.get_dpi_x()
        self.dpi_y = context.get_dpi_y()

    def on_paginate(self, operation, context):
        """Paginate the whole document in chunks."""
        layout = context.create_pango_layout()

        finished = self.paginate(
            layout, self.page_width, self.page_height, self.dpi_x, self.dpi_y
        )
        # update page number
        operation.set_n_pages(len(self._pages))

        # start preview if needed
        if finished and self.preview:
            self.preview.start()

        return finished

    def on_draw_page(self, operation, context, page_nr):
        """Draw the requested page."""
        cr = context.get_cairo_context()
        layout = context.create_pango_layout()
        width = round(context.get_width())
        height = round(context.get_height())
        dpi_x = context.get_dpi_x()
        dpi_y = context.get_dpi_y()

        self.draw_page(page_nr, cr, layout, width, height, dpi_x, dpi_y)

    def on_preview(self, operation, preview, context, parent):
        """Implement custom print preview functionality."""
        ##if constfunc.win()':
        ##return False

        self.preview = PrintPreview(operation, preview, context, parent)

        # give a dummy cairo context to Gtk.PrintContext,
        # PrintPreview will update it with the real one
        try:
            width = int(round(context.get_width()))
        except ValueError:
            width = 0
        try:
            height = int(round(context.get_height()))
        except ValueError:
            height = 0
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
        cr = cairo.Context(surface)
        context.set_cairo_context(cr, PRINTER_DPI, PRINTER_DPI)

        return True



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2006  Donald N. Allingham
#               2009       Gary Burton
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# internationalization
#
# -------------------------------------------------------------------------

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from ..views.treemodels import RepositoryModel
from .baseselector import BaseSelector
from gramps.gen.const import URL_MANUAL_SECT2

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------


# -------------------------------------------------------------------------
#
# SelectRepository
#
# -------------------------------------------------------------------------
class SelectRepository(BaseSelector):
    def _local_init(self):
        """
        Perform local initialisation for this class
        """
        self.setup_configs("interface.repo-sel", 600, 450)

    def get_window_title(self):
        return _("Select Repository")

    def get_model_class(self):
        return RepositoryModel

    def get_column_titles(self):
        return [
            (_("Title"), 350, BaseSelector.TEXT, 0),
            (_("ID"), 75, BaseSelector.TEXT, 1),
            (_("Last Change"), 150, BaseSelector.TEXT, 14),
        ]

    def get_from_handle_func(self):
        return self.db.get_repository_from_handle

    def get_config_name(self):
        return __name__

    WIKI_HELP_PAGE = URL_MANUAL_SECT2
    WIKI_HELP_SEC = _("Select_Repository_selector", "manual")



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# test/GrampsLogger/ErrorReportAssistant_Test.py

import unittest
import logging
import sys
import os

log = logging.getLogger("Gramps.Tests.GrampsLogger")
import gramps.gen.const as const

rootDir = os.path.join(os.path.dirname(__file__), "../../gramps")
sys.path.append(os.path.join(rootDir, "test"))
try:
    from guitest.gtktest import GtkTestCase

    TestCaseBase = GtkTestCase
    log.info("Using guitest")
except:
    TestCaseBase = unittest.TestCase

sys.path.append(rootDir)
sys.path.append(os.path.join(rootDir, "GrampsLogger"))

from gramps.gui.logger import RotateHandler, _errorreportassistant


class ErrorReportAssistantTest(TestCaseBase):  # type: ignore
    """Test the ErrorReportAssistant."""

    def test_buffer_recall(self):
        """Test that simple recall of messages works."""

        rh = RotateHandler(10)
        l = logging.getLogger("ErrorReportAssistantTest")
        l.setLevel(logging.DEBUG)

        l.addHandler(rh)
        l.info("info message")

        # Comment this out because there is noone to close the dialogue
        #         error_detail="Test error"
        #         ass = _errorreportassistant.ErrorReportAssistant(error_detail=error_detail,
        #                                                                rotate_handler=rh)
        #
        #         assert ass._error_detail == error_detail

        l.removeHandler(rh)


def testSuite():
    suite = unittest.TestLoader().loadTestsFromTestCase(ErrorReportAssistantTest)
    return suite


if __name__ == "__main__":
    unittest.TextTestRunner().run(testSuite())



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2016 Tom Samstag
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Unittest that tests repository-specific filter rules
"""
import unittest
import os

from ....db.utils import import_as_dict
from ....filters import GenericFilterFactory
from ....const import DATA_DIR
from ....user import User

from ..repository import (
    AllRepos,
    HasIdOf,
    RegExpIdOf,
    HasNoteRegexp,
    HasReferenceCountOf,
    RepoPrivate,
    ChangedSince,
    MatchesNameSubstringOf,
    HasTag,
)

TEST_DIR = os.path.abspath(os.path.join(DATA_DIR, "tests"))
EXAMPLE = os.path.join(TEST_DIR, "example.gramps")
GenericRepositoryFilter = GenericFilterFactory("Repository")


class BaseTest(unittest.TestCase):
    """
    Repository rule tests.
    """

    @classmethod
    def setUpClass(cls):
        """
        Import example database.
        """
        cls.db = import_as_dict(EXAMPLE, User())

    def filter_with_rule(self, rule):
        """
        Apply a filter with the given rule.
        """
        filter_ = GenericRepositoryFilter()
        filter_.add_rule(rule)
        results = filter_.apply(self.db)
        return set(results)

    def test_allrepos(self):
        """
        Test AllRepos rule.
        """
        rule = AllRepos([])
        self.assertEqual(
            len(self.filter_with_rule(rule)), self.db.get_number_of_repositories()
        )

    def test_hasidof(self):
        """
        Test HasIdOf rule.
        """
        rule = HasIdOf(["R0000"])
        self.assertEqual(self.filter_with_rule(rule), set(["b39fe38593f3f8c4f12"]))

    def test_regexpidof(self):
        """
        Test RegExpIdOf rule.
        """
        rule = RegExpIdOf(["R000."], use_regex=True)
        self.assertEqual(
            self.filter_with_rule(rule),
            set(["a701ead12841521cd4d", "a701e99f93e5434f6f3", "b39fe38593f3f8c4f12"]),
        )

    def test_hasnoteregexp(self):
        """
        Test HasNoteRegexp rule.
        """
        rule = HasNoteRegexp(["."], use_regex=True)
        self.assertEqual(
            self.filter_with_rule(rule),
            set(["a701ead12841521cd4d", "b39fe38593f3f8c4f12"]),
        )

    def test_hasreferencecountof(self):
        """
        Test HasReferenceCountOf rule.
        """
        rule = HasReferenceCountOf(["greater than", "1"])
        self.assertEqual(self.filter_with_rule(rule), set(["a701e99f93e5434f6f3"]))

    def test_repoprivate(self):
        """
        Test RepoPrivate rule.
        """
        rule = RepoPrivate([])
        self.assertEqual(self.filter_with_rule(rule), set([]))

    def test_changedsince(self):
        """
        Test ChangedSince rule.
        """
        rule = ChangedSince(["2010-01-01", "2016-01-01"])
        self.assertEqual(self.filter_with_rule(rule), set(["a701e99f93e5434f6f3"]))

    def test_matchesnamesubstringof(self):
        """
        Test MatchesNameSubstringOf rule.
        """
        rule = MatchesNameSubstringOf(["Martha"])
        self.assertEqual(self.filter_with_rule(rule), set(["a701ead12841521cd4d"]))

    def test_hastag(self):
        """
        Test HasTag rule.
        """
        rule = HasTag(["ToDo"])
        self.assertEqual(self.filter_with_rule(rule), set())


if __name__ == "__main__":
    unittest.main()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2011       Tim G L Lyons, Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
CitationListModel class for Gramps.
"""

# -------------------------------------------------------------------------
#
# python modules
#
# -------------------------------------------------------------------------
import logging

log = logging.getLogger(".")
LOG = logging.getLogger(".citation")

# -------------------------------------------------------------------------
#
# GNOME/GTK modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .flatbasemodel import FlatBaseModel
from .citationbasemodel import CitationBaseModel


# -------------------------------------------------------------------------
#
# CitationListModel
#
# -------------------------------------------------------------------------
class CitationListModel(CitationBaseModel, FlatBaseModel):
    """
    Flat citation model.  (Original code in CitationBaseModel).
    """

    def __init__(
        self,
        db,
        uistate,
        scol=0,
        order=Gtk.SortType.ASCENDING,
        search=None,
        skip=set(),
        sort_map=None,
    ):
        self.map = db.get_raw_citation_data
        self.gen_cursor = db.get_citation_cursor
        self.fmap = [
            self.citation_page,
            self.citation_id,
            self.citation_date,
            self.citation_confidence,
            self.citation_private,
            self.citation_tags,
            self.citation_change,
            self.citation_src_title,
            self.citation_src_id,
            self.citation_src_auth,
            self.citation_src_abbr,
            self.citation_src_pinfo,
            self.citation_src_private,
            self.citation_src_chan,
            self.citation_tag_color,
        ]
        self.smap = [
            self.citation_page,
            self.citation_id,
            self.citation_sort_date,
            self.citation_sort_confidence,
            self.citation_private,
            self.citation_tags,
            self.citation_sort_change,
            self.citation_src_title,
            self.citation_src_id,
            self.citation_src_auth,
            self.citation_src_abbr,
            self.citation_src_pinfo,
            self.citation_src_private,
            self.citation_src_sort_change,
            self.citation_tag_color,
        ]
        FlatBaseModel.__init__(
            self, db, uistate, scol, order, search=search, skip=skip, sort_map=sort_map
        )

    def destroy(self):
        """
        Unset all elements that can prevent garbage collection
        """
        self.db = None
        self.gen_cursor = None
        self.map = None
        self.fmap = None
        self.smap = None
        FlatBaseModel.destroy(self)

    def color_column(self):
        """
        Return the color column.
        """
        return 14

    def on_get_n_columns(self):
        return len(self.fmap) + 1



# -*- python -*-
# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011-2016       Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import time
import logging
from math import pi as PI
from gi.repository import GObject
from gi.repository import Gdk


# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
_LOG = logging.getLogger("maps.markerlayer")

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------

# -------------------------------------------------------------------------
#
# Gramps Modules
#
# -------------------------------------------------------------------------

# -------------------------------------------------------------------------
#
# osmGpsMap
#
# -------------------------------------------------------------------------

try:
    import gi

    gi.require_version("OsmGpsMap", "1.0")
    from gi.repository import OsmGpsMap as osmgpsmap
except:
    raise


class MarkerLayer(GObject.GObject, osmgpsmap.MapLayer):
    """
    This is the layer used to display the markers.
    """

    def __init__(self):
        """
        Initialize the layer
        """
        GObject.GObject.__init__(self)
        self.markers = []
        self.max_references = 0
        self.max_places = 0
        self.nb_ref_by_places = 0
        self.max_value = 0
        self.min_value = 9999

    def clear_markers(self):
        """
        reset the layer attributes.
        """
        self.markers = []
        self.max_references = 0
        self.max_places = 0
        self.nb_ref_by_places = 0
        self.max_value = 0
        self.min_value = 9999

    def add_marker(self, points, image, count, color=None):
        """
        Add a marker.
        Set minimum value, maximum value for the markers
        Set the average value too.
        We calculate that here, to minimize the overhead at markers drawing
        """
        self.markers.append((points, image, count, color))
        self.max_references += count
        self.max_places += 1
        if count > self.max_value:
            self.max_value = count
        if count < self.min_value:
            self.min_value = count
        self.nb_ref_by_places = self.max_references / self.max_places

    def do_draw(self, gpsmap, ctx):
        """
        Draw all markers here. Calculate where to draw the marker.
        Depending of the average, minimum and maximum value, resize the marker.
        We use cairo to resize the marker.
        """
        max_interval = self.max_value - self.nb_ref_by_places
        min_interval = self.nb_ref_by_places - self.min_value
        if max_interval <= 0:  # This to avoid divide by zero
            max_interval = 0.01
        if min_interval <= 0:  # This to avoid divide by zero
            min_interval = 0.01
        _LOG.debug(
            "%s",
            time.strftime("start drawing   : " "%a %d %b %Y %H:%M:%S", time.gmtime()),
        )
        for marker in self.markers:
            # the icon size in 48, so the standard icon size is 0.6 * 48 = 28.8
            size = 0.6
            mark = float(marker[2])
            if mark > self.nb_ref_by_places or max_interval > 3:
                # at maximum, we'll have an icon size = (0.6 + 0.2) * 48 = 38.4
                size += 0.2 * ((mark - self.nb_ref_by_places) / max_interval)
            else:
                # at minimum, we'll have an icon size = (0.6 - 0.2) * 48 = 19.2
                size -= 0.2 * ((self.nb_ref_by_places - mark) / min_interval)

            conv_pt = osmgpsmap.MapPoint.new_degrees(
                float(marker[0][0]), float(marker[0][1])
            )
            coord_x, coord_y = gpsmap.convert_geographic_to_screen(conv_pt)
            if marker[3] is None:
                # We use the standard icons.
                ctx.save()
                ctx.translate(coord_x, coord_y)
                ctx.scale(size, size)
                # below, we try to place exactly the marker depending on its
                # size. The left top corner of the image is set to the
                # coordinates. The tip of the pin which should be at the marker
                # position is at 3/18 of the width and to the height of the
                # image. So we shift the image position.
                pos_y = -int(48 * size + 0.5) - 10
                pos_x = -int((48 * size) / 6 + 0.5) - 10
                ctx.set_source_surface(marker[1], pos_x, pos_y)
                ctx.paint()
                ctx.restore()
            else:
                # We use colored icons.
                draw_marker(ctx, float(coord_x), float(coord_y), size, marker[3][1])
        _LOG.debug(
            "%s",
            time.strftime("end drawing     : " "%a %d %b %Y %H:%M:%S", time.gmtime()),
        )

    def do_render(self, gpsmap):
        """
        render the layer
        """
        pass

    def do_busy(self):
        """
        set the layer busy
        """
        return False

    def do_button_press(self, gpsmap, gdkeventbutton):
        """
        When we press a button.
        """
        dummy_map = gpsmap
        dummy_evt = gdkeventbutton
        return False


GObject.type_register(MarkerLayer)


def draw_marker(ctx, x01, y01, size, color):
    width = 48.0 * size
    height = width / 2
    rgba = Gdk.RGBA()
    rgba.parse(color)
    fill_color = (
        rgba.red,
        rgba.green,
        rgba.blue,
        1.0,  # transparency
    )
    stroke_color = (1.0, 0.0, 0.0, 0.5)

    ctx.move_to(x01, y01)
    ctx.line_to(x01 + width / 6, y01 - height * 2)
    ctx.line_to(x01 - width / 6, y01 - height * 2)
    ctx.close_path()
    ctx.set_line_width(2.0)
    ctx.set_source_rgba(*fill_color)
    ctx.fill_preserve()
    ctx.set_source_rgba(*stroke_color)
    ctx.stroke()

    ctx.save()
    ctx.translate(x01, y01 - height * 2 - width / 4)
    ctx.scale(width / 2, height / 2)
    ctx.arc(0.0, 0.0, 1.0, 0.0, 2 * PI)
    ctx.restore()
    ctx.set_line_width(2.0)
    ctx.set_source_rgba(*fill_color)
    ctx.fill_preserve()
    ctx.set_source_rgba(*stroke_color)
    ctx.stroke()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Parses the lds.xml file to build the temple/code maps
"""

from ..const import DATA_DIR
import os
import logging

from xml.parsers.expat import ParserCreate
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

LOG = logging.getLogger(".")


class LdsTemples:
    """
    Parsing class for the LDS temples file
    """

    def __init__(self):
        """
        Parses the lds.xml file to load the LDS temple code to name
        maps
        """
        self.__temple_codes = {}
        self.__temple_to_abrev = {}
        self.__current_temple = ""
        self.__tlist = []

        lds_filename = os.path.expanduser(os.path.join(DATA_DIR, "lds.xml"))

        try:
            parser = ParserCreate()
            parser.StartElementHandler = self.__start_element
            parser.EndElementHandler = self.__end_element
            parser.CharacterDataHandler = self.__characters
            with open(lds_filename, "rb") as xml_file:
                parser.ParseFile(xml_file)
        except Exception as msg:
            LOG.error(str(msg))

    def is_valid_code(self, code):
        """
        returns True if the code is a valid LDS temple code according
        to the lds.xml file
        """
        return self.__temple_to_abrev.get(code) is not None

    def is_valid_name(self, name):
        """
        returns True if the name matches a temple name (not code) in
        the lds.xml file
        """
        return self.__temple_codes.get(name) is not None

    def code(self, name):
        """
        returns the LDS Temple code that corresponds to the name
        """
        return self.__temple_codes.get(name, _("Unknown"))

    def name(self, code):
        """
        returns the name associated with the LDS Temple code
        """
        return self.__temple_to_abrev.get(code, _("Unknown"))

    def name_code_data(self):
        """
        returns a list of temple codes, temple name tuples
        """
        return sorted(
            [(code, name) for name, code in self.__temple_codes.items()],
            key=lambda v: v[1],
        )

    def __start_element(self, tag, attrs):
        """
        XML parsing function that is called when an XML element is first found
        """
        self.__tlist = []
        if tag == "temple":
            self.__current_temple = attrs.get("name")

    def __end_element(self, tag):
        """
        XML parsing function that is called when an XML element is closed
        """

        text = "".join(self.__tlist)

        if tag == "code":
            if self.__temple_codes.get(self.__current_temple) is None:
                self.__temple_codes[self.__current_temple] = text
            self.__temple_to_abrev[text] = self.__current_temple

    def __characters(self, data):
        """
        XML parsing function that collects text data
        """
        self.__tlist.append(data)


TEMPLES = LdsTemples()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2014-2017  Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Class handling displaying of places.
"""

# ---------------------------------------------------------------
#
# Python imports
#
# ---------------------------------------------------------------
import os
import xml.dom.minidom

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..const import PLACE_FORMATS, GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from ..config import config
from ..utils.location import get_location_list
from ..lib import PlaceType


# -------------------------------------------------------------------------
#
# PlaceFormat class
#
# -------------------------------------------------------------------------
class PlaceFormat:
    def __init__(self, name, levels, language, street, reverse):
        self.name = name
        self.levels = levels
        self.language = language
        self.street = street
        self.reverse = reverse


# -------------------------------------------------------------------------
#
# PlaceDisplay class
#
# -------------------------------------------------------------------------
class PlaceDisplay:
    def __init__(self):
        self.place_formats = []
        self.default_format = config.get("preferences.place-format")
        if os.path.exists(PLACE_FORMATS):
            try:
                self.load_formats()
                return
            except BaseException:
                print(_("Error in '%s' file: cannot load.") % PLACE_FORMATS)
        pf = PlaceFormat(_("Full"), ":", "", 0, False)
        self.place_formats.append(pf)

    def display_event(self, db, event, fmt=-1):
        if not event:
            return ""
        place_handle = event.get_place_handle()
        if place_handle:
            place = db.get_place_from_handle(place_handle)
            return self.display(db, place, event.get_date_object(), fmt)
        else:
            return ""

    def display(self, db, place, date=None, fmt=-1):
        if not place:
            return ""
        if not config.get("preferences.place-auto"):
            return place.title
        else:
            if fmt == -1:
                fmt = config.get("preferences.place-format")
            pf = self.place_formats[fmt]
            lang = pf.language
            all_places = get_location_list(db, place, date, lang)

            # Apply format string to place list
            index = _find_populated_place(all_places)
            places = []
            for slice in pf.levels.split(","):
                parts = slice.split(":")
                if len(parts) == 1:
                    offset = _get_offset(parts[0], index)
                    if offset is not None:
                        try:
                            places.append(all_places[offset])
                        except IndexError:
                            pass
                elif len(parts) == 2:
                    start = _get_offset(parts[0], index)
                    end = _get_offset(parts[1], index)
                    if start is None:
                        places.extend(all_places[:end])
                    elif end is None:
                        places.extend(all_places[start:])
                    else:
                        places.extend(all_places[start:end])

            if pf.street:
                types = [item[1] for item in places]
                try:
                    idx = types.index(PlaceType.NUMBER)
                except ValueError:
                    idx = None
                if idx is not None and len(places) > idx + 1:
                    if pf.street == 1:
                        combined = (
                            places[idx][0] + " " + places[idx + 1][0],
                            places[idx + 1][1],
                        )
                    else:
                        combined = (
                            places[idx + 1][0] + " " + places[idx][0],
                            places[idx + 1][1],
                        )
                    places = places[:idx] + [combined] + places[idx + 2 :]

            names = [item[0] for item in places]
            if pf.reverse:
                names.reverse()

            # TODO for Arabic, should the next line's comma be translated?
            return ", ".join(names)

    def get_formats(self):
        return self.place_formats

    def set_formats(self, formats):
        self.place_formats = formats

    def load_formats(self):
        dom = xml.dom.minidom.parse(PLACE_FORMATS)
        top = dom.getElementsByTagName("place_formats")

        for fmt in top[0].getElementsByTagName("format"):
            name = fmt.attributes["name"].value
            levels = fmt.attributes["levels"].value
            language = fmt.attributes["language"].value
            street = int(fmt.attributes["street"].value)
            reverse = fmt.attributes["reverse"].value == "True"
            pf = PlaceFormat(name, levels, language, street, reverse)
            self.place_formats.append(pf)

        dom.unlink()

    def save_formats(self):
        doc = xml.dom.minidom.Document()
        place_formats = doc.createElement("place_formats")
        doc.appendChild(place_formats)
        for fmt in self.place_formats:
            node = doc.createElement("format")
            place_formats.appendChild(node)
            node.setAttribute("name", fmt.name)
            node.setAttribute("levels", fmt.levels)
            node.setAttribute("language", fmt.language)
            node.setAttribute("street", str(fmt.street))
            node.setAttribute("reverse", str(fmt.reverse))
        with open(PLACE_FORMATS, "w", encoding="utf-8") as f_d:
            doc.writexml(f_d, addindent="  ", newl="\n", encoding="utf-8")


def _get_offset(value, index):
    if index is not None and value.startswith("p"):
        try:
            offset = int(value[1:])
        except ValueError:
            offset = 0
        offset += index
    else:
        try:
            offset = int(value)
        except ValueError:
            offset = None
    return offset


def _find_populated_place(places):
    populated_place = None
    for index, item in enumerate(places):
        if int(item[1]) in [
            PlaceType.HAMLET,
            PlaceType.VILLAGE,
            PlaceType.TOWN,
            PlaceType.CITY,
        ]:
            populated_place = index
    return populated_place


displayer = PlaceDisplay()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python classes
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import GObject
from gi.repository import GLib

# -------------------------------------------------------------------------
#
# Gramps classes
#
# -------------------------------------------------------------------------
from gramps.gen.errors import WindowActiveError
from gramps.gen.lib import Note
from ...dbguielement import DbGUIElement
from ...selectors import SelectorFactory
from .notemodel import NoteModel
from .embeddedlist import EmbeddedList, TEXT_COL, MARKUP_COL, ICON_COL
from ...ddtargets import DdTargets


# -------------------------------------------------------------------------
#
# NoteTab
#
# -------------------------------------------------------------------------
class NoteTab(EmbeddedList, DbGUIElement):
    """
    Note List display tab for edit dialogs.

    Derives from the EmbeddedList class.
    """

    _HANDLE_COL = 3
    _DND_TYPE = DdTargets.NOTE_LINK

    _MSG = {
        "add": _("Create and add a new note"),
        "del": _("Remove the existing note"),
        "edit": _("Edit the selected note"),
        "share": _("Add an existing note"),
        "up": _("Move the selected note upwards"),
        "down": _("Move the selected note downwards"),
    }

    # index = column in model. Value =
    #  (name, sortcol in model, width, markup/text, weigth_col
    _column_names = [
        (_("Type"), 0, 100, TEXT_COL, -1, None),
        (_("Preview"), 1, 200, TEXT_COL, -1, None),
        (_("Private"), 2, 30, ICON_COL, -1, "gramps-lock"),
    ]

    def __init__(
        self,
        dbstate,
        uistate,
        track,
        data,
        config_key,
        callertitle=None,
        notetype=None,
    ):
        self.data = data
        self.callertitle = callertitle
        self.notetype = notetype
        EmbeddedList.__init__(
            self,
            dbstate,
            uistate,
            track,
            _("_Notes"),
            NoteModel,
            config_key,
            share_button=True,
            move_buttons=True,
        )
        DbGUIElement.__init__(self, dbstate.db)
        self.callman.register_handles({"note": self.data})

    def _connect_db_signals(self):
        """
        Implement base class DbGUIElement method
        """
        # note: note-rebuild closes the editors, so no need to connect to it
        self.callman.register_callbacks(
            {
                "note-delete": self.note_delete,  # delete a note we track
                "note-update": self.note_update,  # change a note we track
            }
        )
        self.callman.connect_all(keys=["note"])

    def get_editor(self):
        pass

    def get_user_values(self):
        return []

    def get_data(self):
        """
        Return the data associated with display tab
        """
        return self.data

    def column_order(self):
        """
        Return the column order of the columns in the display tab.
        """
        return ((1, 2), (1, 0), (1, 1))

    def add_button_clicked(self, obj):
        """
        Create a new Note instance and call the EditNote editor with the new
        note.

        Called when the Add button is clicked.
        If the window already exists (WindowActiveError), we ignore it.
        This prevents the dialog from coming up twice on the same object.
        """
        note = Note()
        if self.notetype:
            note.set_type(self.notetype)
        try:
            from .. import EditNote

            EditNote(
                self.dbstate,
                self.uistate,
                self.track,
                note,
                self.add_callback,
                self.callertitle,
                extratype=[self.notetype],
            )
        except WindowActiveError:
            pass

    def add_callback(self, name):
        """
        Called to update the screen when a new note is added
        """
        data = self.get_data()
        data.append(name)
        self.callman.register_handles({"note": [name]})
        self.changed = True
        self.rebuild()
        GLib.idle_add(self.tree.scroll_to_cell, len(data) - 1)

    def edit_button_clicked(self, obj):
        """
        Get the selected Note instance and call the EditNote editor with the
        note.

        Called when the Edit button is clicked.
        If the window already exists (WindowActiveError), we ignore it.
        This prevents the dialog from coming up twice on the same object.
        """
        handle = self.get_selected()
        if handle:
            note = self.dbstate.db.get_note_from_handle(handle)
            try:
                from .. import EditNote

                EditNote(
                    self.dbstate,
                    self.uistate,
                    self.track,
                    note,
                    callertitle=self.callertitle,
                    extratype=[self.notetype],
                )
            except WindowActiveError:
                pass

    def share_button_clicked(self, obj):
        SelectNote = SelectorFactory("Note")

        sel = SelectNote(self.dbstate, self.uistate, self.track)
        note = sel.run()
        if note:
            self.add_callback(note.handle)

    def get_icon_name(self):
        """
        Return the icon name associated with the display tab
        """
        return "gramps-notes"

    def note_delete(self, del_note_handle_list):
        """
        Outside of this tab note objects have been deleted. Check if tab
        and object must be changed.
        Note: delete of object will cause reference on database to be removed,
            so this method need not do this
        """
        rebuild = False
        for handle in del_note_handle_list:
            while self.data.count(handle) > 0:
                self.data.remove(handle)
                rebuild = True
        if rebuild:
            self.rebuild()

    def note_update(self, upd_note_handle_list):
        """
        Outside of this tab note objects have been updated. Check if tab
        and object must be updated.
        """
        for handle in upd_note_handle_list:
            if handle in self.data:
                self.rebuild()
                break



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2022       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
A thumbnailer for images and sub-sections of images.
"""

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import logging

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import GdkPixbuf

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import THUMBSCALE, THUMBSCALE_LARGE, SIZE_LARGE
from gramps.gen.plug import Thumbnailer

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
LOG = logging.getLogger(".thumbnail")


class ImageThumb(Thumbnailer):
    def is_supported(self, mime_type):
        if mime_type:
            return mime_type.startswith("image")
        else:
            return False

    def run(self, mime_type, src_file, dest_file, size, rectangle):
        # build a thumbnail by scaling the image using GTK's built in
        # routines.
        try:
            pixbuf = GdkPixbuf.Pixbuf.new_from_file(src_file)
            width = pixbuf.get_width()
            height = pixbuf.get_height()

            if rectangle is not None:
                upper_x = min(rectangle[0], rectangle[2]) / 100.0
                lower_x = max(rectangle[0], rectangle[2]) / 100.0
                upper_y = min(rectangle[1], rectangle[3]) / 100.0
                lower_y = max(rectangle[1], rectangle[3]) / 100.0
                sub_x = int(upper_x * width)
                sub_y = int(upper_y * height)
                sub_width = int((lower_x - upper_x) * width)
                sub_height = int((lower_y - upper_y) * height)
                if sub_width > 0 and sub_height > 0:
                    pixbuf = pixbuf.new_subpixbuf(sub_x, sub_y, sub_width, sub_height)
                    width = sub_width
                    height = sub_height

            if size == SIZE_LARGE:
                thumbscale = THUMBSCALE_LARGE
            else:
                thumbscale = THUMBSCALE
            scale = thumbscale / (float(max(width, height)))

            scaled_width = int(width * scale)
            scaled_height = int(height * scale)

            pixbuf = pixbuf.scale_simple(
                scaled_width, scaled_height, GdkPixbuf.InterpType.BILINEAR
            )
            pixbuf.savev(dest_file, "png", "", "")
            return True
        except Exception as err:
            LOG.warning("Error scaling image down: %s", str(err))
            return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2008       Zsolt Foldvari
# Copyright (C) 2013       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2017,2024  Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"Provide formatting tag definition for StyledText."

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..const import GRAMPS_LOCALE as glocale
from .baseobj import BaseObject
from .styledtexttagtype import StyledTextTagType

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# StyledTextTag
#
# -------------------------------------------------------------------------
class StyledTextTag(BaseObject):
    """Hold formatting information for :py:class:`.StyledText`.

    :py:class:`StyledTextTag` is a container class, it's attributes are
    directly accessed.

    :ivar name: Type (or name) of the tag instance. E.g. 'bold', etc.
    :type name: :py:class:`.StyledTextTagType` instace
    :ivar value: Value of the tag. E.g. color hex string for font color, etc.
    :type value: str or None
    :ivar ranges: Pointer pairs into the string, where the tag applies.
    :type ranges: list of (int(start), int(end)) tuples.

    """

    def __init__(self, name=None, value=None, ranges=None):
        """Setup initial instance variable values.

        .. note:: Since :py:class:`.GrampsType` supports the instance
                  initialization with several different base types, please note
                  that ``name`` parameter can be int, str, unicode, tuple,
                  or even another :py:class:`.StyledTextTagType` instance.
        """
        self.name = StyledTextTagType(name)
        self.value = value
        if ranges is None:
            self.ranges = []
        else:
            # Current use of StyledTextTag is such that a shallow copy suffices.
            self.ranges = ranges

    def set_object_state(self, attr_dict):
        """
        Set the current object state using information provided in the given
        dictionary.

        We override this method to convert the elements of `ranges` into tuples.
        """
        attr_dict["ranges"] = [tuple(item) for item in attr_dict["ranges"]]
        super().set_object_state(attr_dict)

    def serialize(self):
        """Convert the object to a serialized tuple of data.

        :return: Serialized format of the instance.
        :rtype: tuple

        """
        return (self.name.serialize(), self.value, self.ranges)

    def unserialize(self, data):
        """Convert a serialized tuple of data to an object.

        :param data: Serialized format of instance variables.
        :type data: tuple

        """
        (the_name, self.value, self.ranges) = data

        self.name = StyledTextTagType()
        self.name.unserialize(the_name)
        return self

    @classmethod
    def get_schema(cls):
        """
        Returns the JSON Schema for this class.

        :returns: Returns a dict containing the schema.
        :rtype: dict
        """
        return {
            "type": "object",
            "title": _("Tag"),
            "properties": {
                "_class": {"enum": [cls.__name__]},
                "name": StyledTextTagType.get_schema(),
                "value": {
                    "type": ["null", "string", "integer"],
                    "title": _("Value"),
                },
                "ranges": {
                    "type": "array",
                    "items": {
                        "type": "array",
                        "items": {"type": "integer"},
                        "minItems": 2,
                        "maxItems": 2,
                    },
                    "title": _("Ranges"),
                },
            },
        }



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Arabic-specific classes for parsing and displaying dates.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import re

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..lib.date import Date
from ._dateparser import DateParser
from ._datedisplay import DateDisplay
from ._datehandler import register_datehandler
from ..const import ARABIC_COMMA


# -------------------------------------------------------------------------
#
# Arabic parser class
#
# -------------------------------------------------------------------------
class DateParserAR(DateParser):
    """
    Convert a text string into a Date object. If the date cannot be
    converted, the text string is assigned.
    """

    # modifiers before the date
    modifier_to_int = {
        "قبل": Date.MOD_BEFORE,
        "قبل": Date.MOD_BEFORE,
        "قبل.": Date.MOD_BEFORE,
        "بعد": Date.MOD_AFTER,
        "بعد": Date.MOD_AFTER,
        "بعد.": Date.MOD_AFTER,
        "حوالي": Date.MOD_ABOUT,
        "حوالي.": Date.MOD_ABOUT,
        "حوالي": Date.MOD_ABOUT,
        "حوالي": Date.MOD_ABOUT,
        "حوالي.": Date.MOD_ABOUT,
        "حوالي": Date.MOD_ABOUT,
        "from": Date.MOD_FROM,
        "to": Date.MOD_TO,
    }

    islamic_to_int = {
        "محرّم": 1,
        "محرّم الحرام": 1,
        "صفر": 2,
        "ربيع الأول": 3,
        "ربيع 1": 3,
        "ربيع الأخير": 4,
        "ربيع الثاني": 4,
        "ربيع الثاني": 4,
        "ربيع الثاني": 4,
        "ربيع الثاني": 4,
        "ربيع 2": 4,
        "جمادى الأولى": 5,
        "جمادى الأول": 5,
        "جمادى 1": 5,
        "جمادى الثانية": 6,
        "جمادى الأخير": 6,
        "جمادى الثاني": 6,
        "جمادى 2": 5,
        "رجب": 7,
        "شعبان": 8,
        "شعبان": 8,
        "رمضان": 9,
        "رمضان": 9,
        "شوال": 10,
        "ذو القعدة": 11,
        "ذو القعدة": 11,
        "ذو القعدة": 11,
        "ذو الحجة": 12,
        "ذو الحجة": 12,
        "ذو الحجة": 12,
    }

    bce = [
        "قبل الميلاد",
        "قبل الميلاد",
        "قبل الميلاد",
        "قبل الميلاد",
        "قبل الميلاد",
        "قبل الميلاد",
    ]

    calendar_to_int = {
        "غريغوري": Date.CAL_GREGORIAN,
        "غريغوري": Date.CAL_GREGORIAN,
        "يوليوسي": Date.CAL_JULIAN,
        "يوليوسي": Date.CAL_JULIAN,
        "عبري": Date.CAL_HEBREW,
        "عبري": Date.CAL_HEBREW,
        "إسلامي": Date.CAL_ISLAMIC,
        "إسلامي": Date.CAL_ISLAMIC,
        "فرنسي": Date.CAL_FRENCH,
        "فرنسي جمهوري": Date.CAL_FRENCH,
        "فرنسي": Date.CAL_FRENCH,
        "فارسي": Date.CAL_PERSIAN,
        "فارسي": Date.CAL_PERSIAN,
        "سويدي": Date.CAL_SWEDISH,
        "سويدي": Date.CAL_SWEDISH,
    }

    quality_to_int = {
        "متوقع": Date.QUAL_ESTIMATED,
        "متوقع.": Date.QUAL_ESTIMATED,
        "متوقع": Date.QUAL_ESTIMATED,
        "محسوب.": Date.QUAL_CALCULATED,
        "محسوب": Date.QUAL_CALCULATED,
        "محسوب": Date.QUAL_CALCULATED,
    }

    def init_strings(self):
        """
        This method compiles regular expression strings for matching dates.
        """
        DateParser.init_strings(self)
        _span_1 = ["من"]
        _span_2 = ["إلى"]
        _range_1 = ["بين"]
        _range_2 = ["و"]
        self._span = re.compile(
            r"(%s)\s+(?P<start>.+)\s+(%s)\s+(?P<stop>.+)"
            % ("|".join(_span_1), "|".join(_span_2)),
            re.IGNORECASE,
        )
        self._range = re.compile(
            r"(%s)\s+(?P<start>.+)\s+(%s)\s+(?P<stop>.+)"
            % ("|".join(_range_1), "|".join(_range_2)),
            re.IGNORECASE,
        )


# -------------------------------------------------------------------------
#
# Arabic display
#
# -------------------------------------------------------------------------
class DateDisplayAR(DateDisplay):
    """
    Arabic language date display class.
    """

    # this is used to display the 12 gregorian months
    long_months = (
        "",
        "كانون الثاني",
        "شباط",
        "آذار",
        "نيسان",
        "أيار",
        "حزيران",
        "تموز",
        "آب",
        "أيلول",
        "تشرين الأول",
        "تشرين الثاني",
        "كانون الأول",
    )

    short_months = (
        "",
        "كانون2",
        "شباط",
        "آذار",
        "نيسان",
        "أيار",
        "حزيران",
        "تموز",
        "آب",
        "أيلول",
        "تشرين1",
        "تشرين2",
        "كانون1",
    )

    islamic = (
        "",
        "محرّم",
        "صفر",
        "ربيع الأول",
        "ربيع الثاني",
        "جمادى الأولى",
        "جمادى الثانية",
        "رجب",
        "شعبان",
        "رمضان",
        "شوال",
        "ذو القعدة",
        "ذو الحجة",
    )

    formats = (
        "YYYY-MM-DD (قياسي)",
        "عددي",
        "شهر يوم, سنة",
        "شهر يوم, سنة",
        "يوم شهر سنة",
        "يوم شهر سنة",
    )
    # this must agree with DateDisplayEn's "formats" definition
    # (since no locale-specific _display_gregorian exists, here)

    calendar = ("", "يوليوسي", "عبري", "فرنسي", "فارسي", "إسلامي", "سويدي")

    _mod_str = ("", "قبل ", "بعد ", "حوالي ", "", "", "")

    _qual_str = ("", "متوقع ", "محسوب ")

    _bce_str = "%s قبل الميلاد."

    def display(self, date):
        """
        Return a text string representing the date.
        """
        mod = date.get_modifier()
        cal = date.get_calendar()
        qual = date.get_quality()
        start = date.get_start_date()
        newyear = date.get_new_year()

        qual_str = self._qual_str[qual]

        if mod == Date.MOD_TEXTONLY:
            return date.get_text()
        elif start == Date.EMPTY:
            return ""
        elif mod == Date.MOD_SPAN:
            d1 = self.display_cal[cal](start)
            d2 = self.display_cal[cal](date.get_stop_date())
            scal = self.format_extras(cal, newyear)
            return "%s%s %s %s %s%s" % (qual_str, "من", d1, "إلى", d2, scal)
        elif mod == Date.MOD_RANGE:
            d1 = self.display_cal[cal](start)
            d2 = self.display_cal[cal](date.get_stop_date())
            scal = self.format_extras(cal, newyear)
            return "%s%s %s %s %s%s" % (qual_str, "بين", d1, "و", d2, scal)
        else:
            text = self.display_cal[date.get_calendar()](start)
            scal = self.format_extras(cal, newyear)
            return "%s%s%s%s" % (qual_str, self._mod_str[mod], text, scal)

    def dd_dformat01(self, date_val):
        """
        numerical -- for Arabic dates
        """
        value = DateDisplay.dd_dformat01(self, date_val)
        return value.replace(",", ARABIC_COMMA)

    def dd_dformat02(self, date_val, inflect, long_months):
        """
        month_name day, year -- for Arabic dates
        """
        value = DateDisplay.dd_dformat02(self, date_val, inflect, long_months)
        return value.replace(",", ARABIC_COMMA)

    def dd_dformat03(self, date_val, inflect, short_months):
        """
        month_abbreviation day, year -- for Arabic dates
        """
        value = DateDisplay.dd_dformat03(self, date_val, inflect, short_months)
        return value.replace(",", ARABIC_COMMA)


# -------------------------------------------------------------------------
#
# Register classes
#
# -------------------------------------------------------------------------
register_datehandler(
    ("ar_EG", "ar_AR", "ar", "Arabic", "arabic", ("%d %b, %Y",)),
    DateParserAR,
    DateDisplayAR,
)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009 Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
from gramps.gen.plug._pluginreg import *
from gramps.gen.const import URL_MANUAL_PAGE, GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

MODULE_VERSION = "6.0"
REPORTS_HELP = URL_MANUAL_PAGE + "_-_Reports#Web_Pages"

# ------------------------------------------------------------------------
#
# Narrated Web Site
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "navwebpage"
plg.name = _("Narrated Web Site")
plg.description = _(
    "Produces web (HTML) pages for individuals, or a set of " "individuals"
)
plg.version = "2.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "narrativeweb.py"
plg.ptype = REPORT
plg.authors = ["Donald N. Allingham", "Rob G. Healey", "Serge Noiraud"]
plg.authors_email = ["don@gramps-project.org", "serge.noiraud@free.fr"]
plg.category = CATEGORY_WEB
plg.reportclass = "NavWebReport"
plg.optionclass = "NavWebOptions"
plg.report_modes = [REPORT_MODE_GUI, REPORT_MODE_CLI]
plg.help_url = REPORTS_HELP

# ------------------------------------------------------------------------
#
# Web Calendar
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "WebCal"
plg.name = _("Web Calendar")
plg.description = _("Produces web (HTML) calendars.")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "webcal.py"
plg.ptype = REPORT
plg.authors = ["Thom Sturgill", "Rob G. Healey"]
plg.authors_email = ["thsturgill@yahoo.com", "robhealey1@gmail.com"]
plg.category = CATEGORY_WEB
plg.reportclass = "WebCalReport"
plg.optionclass = "WebCalOptions"
plg.report_modes = [REPORT_MODE_GUI, REPORT_MODE_CLI]
plg.help_url = REPORTS_HELP



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2010       Michiel D. Nauta
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2017       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Citation object for Gramps.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import logging

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..const import GRAMPS_LOCALE as glocale
from .attrbase import SrcAttributeBase
from .citationbase import IndirectCitationBase
from .datebase import DateBase
from .mediabase import MediaBase
from .notebase import NoteBase
from .primaryobj import PrimaryObject
from .tagbase import TagBase

_ = glocale.translation.gettext

LOG = logging.getLogger(".citation")


# -------------------------------------------------------------------------
#
# Citation
#
# -------------------------------------------------------------------------
class Citation(
    MediaBase,
    NoteBase,
    SrcAttributeBase,
    IndirectCitationBase,
    DateBase,
    PrimaryObject,
):
    """
    A record of a citation of a source of information.

    In GEDCOM this is called a SOURCE_CITATION.
    The data provided in the <<SOURCE_CITATION>> structure is source-related
    information specific to the data being cited.
    """

    CONF_VERY_HIGH = 4
    CONF_HIGH = 3
    CONF_NORMAL = 2
    CONF_LOW = 1
    CONF_VERY_LOW = 0

    def __init__(self):
        """Create a new Citation instance."""
        PrimaryObject.__init__(self)
        MediaBase.__init__(self)  #  7
        NoteBase.__init__(self)  #  6
        DateBase.__init__(self)  #  2
        self.source_handle = None  #  5
        self.page = ""  #  3
        self.confidence = Citation.CONF_NORMAL  #  4
        SrcAttributeBase.__init__(self)  #  8

    @classmethod
    def get_schema(cls):
        """
        Returns the JSON Schema for this class.

        :returns: Returns a dict containing the schema.
        :rtype: dict
        """
        # pylint: disable=import-outside-toplevel
        from .date import Date
        from .mediaref import MediaRef
        from .srcattribute import SrcAttribute

        return {
            "type": "object",
            "title": _("Citation"),
            "properties": {
                "_class": {"enum": [cls.__name__]},
                "handle": {
                    "type": "string",
                    "maxLength": 50,
                    "title": _("Handle"),
                },
                "gramps_id": {"type": "string", "title": _("Gramps ID")},
                "date": Date.get_schema(),
                "page": {"type": "string", "title": _("Page")},
                "confidence": {
                    "type": "integer",
                    "minimum": 0,
                    "maximum": 4,
                    "title": _("Confidence"),
                },
                "source_handle": {
                    "type": "string",
                    "maxLength": 50,
                    "title": _("Source"),
                },
                "note_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Notes"),
                },
                "media_list": {
                    "type": "array",
                    "items": MediaRef.get_schema(),
                    "title": _("Media"),
                },
                "attribute_list": {
                    "type": "array",
                    "items": SrcAttribute.get_schema(),
                    "title": _("Source Attributes"),
                },
                "change": {"type": "integer", "title": _("Last changed")},
                "tag_list": {
                    "type": "array",
                    "items": {"type": "string", "maxLength": 50},
                    "title": _("Tags"),
                },
                "private": {"type": "boolean", "title": _("Private")},
            },
        }

    def serialize(self, no_text_date=False):
        """
        Convert the object to a serialized tuple of data.
        """
        return (
            self.handle,  #  0
            self.gramps_id,  #  1
            DateBase.serialize(self, no_text_date),  #  2
            str(self.page),  #  3
            self.confidence,  #  4
            self.source_handle,  #  5
            NoteBase.serialize(self),  #  6
            MediaBase.serialize(self),  #  7
            SrcAttributeBase.serialize(self),  #  8
            self.change,  #  9
            TagBase.serialize(self),  # 10
            self.private,
        )  # 11

    def unserialize(self, data):
        """
        Convert the data held in a tuple created by the serialize method
        back into the data in a Citation structure.
        """
        (
            self.handle,  #  0
            self.gramps_id,  #  1
            date,  #  2
            self.page,  #  3
            self.confidence,  #  4
            self.source_handle,  #  5
            note_list,  #  6
            media_list,  #  7
            srcattr_list,  #  8
            self.change,  #  9
            tag_list,  # 10
            self.private,  # 11
        ) = data

        DateBase.unserialize(self, date)
        NoteBase.unserialize(self, note_list)
        MediaBase.unserialize(self, media_list)
        TagBase.unserialize(self, tag_list)
        SrcAttributeBase.unserialize(self, srcattr_list)
        return self

    def _has_handle_reference(self, classname, handle):
        """
        Return True if the object has reference to a given handle of given
        primary object type.

        :param classname: The name of the primary object class.
        :type classname: str
        :param handle: The handle to be checked.
        :type handle: str
        :returns: Returns whether the object has reference to this handle of
                  this object type.
        :rtype: bool
        """
        if classname == "Note":
            return handle in [ref.ref for ref in self.note_list]
        if classname == "Media":
            return handle in [ref.ref for ref in self.media_list]
        if classname == "Source":
            return handle == self.get_reference_handle()
        return False

    def _remove_handle_references(self, classname, handle_list):
        """
        Remove all references in this object to object handles in the list.

        :param classname: The name of the primary object class.
        :type classname: str
        :param handle_list: The list of handles to be removed.
        :type handle_list: str
        """
        if classname == "Source" and self.get_reference_handle() in handle_list:
            self.set_reference_handle(None)

    def _replace_handle_reference(self, classname, old_handle, new_handle):
        """
        Replace all references to old handle with those to the new handle.

        :param classname: The name of the primary object class.
        :type classname: str
        :param old_handle: The handle to be replaced.
        :type old_handle: str
        :param new_handle: The handle to replace the old one with.
        :type new_handle: str
        """
        if classname == "Source" and self.get_reference_handle() == old_handle:
            self.set_reference_handle(new_handle)

    def get_citation_child_list(self):
        """
        Return the list of child secondary objects that may refer citations.

        :returns: Returns the list of child secondary child objects that may
                  refer citations.
        :rtype: list
        """
        return self.media_list

    def get_text_data_list(self):
        """
        Return the list of all textual attributes of the object.

        :returns: Returns the list of all textual attributes of the object.
        :rtype: list
        """
        return [self.page, self.gramps_id]

    def get_text_data_child_list(self):
        """
        Return the list of child objects that may carry textual data.

        :returns: Returns the list of child objects that may carry textual data.
        :rtype: list
        """
        return self.media_list + self.attribute_list

    def get_note_child_list(self):
        """
        Return the list of child secondary objects that may refer notes.

        :returns: Returns the list of child secondary child objects that may
                  refer notes.
        :rtype: list
        """
        return self.media_list

    def get_handle_referents(self):
        """
        Return the list of child objects which may, directly or through
        their children, reference primary objects.

        :returns: Returns the list of objects referencing primary objects.
        :rtype: list
        """
        return self.media_list

    def get_referenced_handles(self):
        """
        Return the list of (classname, handle) tuples for all directly
        referenced primary objects.

        :returns: List of (classname, handle) tuples for referenced objects.
        :rtype: list
        """
        ret = self.get_referenced_note_handles() + self.get_referenced_tag_handles()
        if self.get_reference_handle():
            ret += [("Source", self.get_reference_handle())]
        return ret

    def merge(self, acquisition):
        """
        Merge the content of acquisition into this citation.

        :param acquisition: The citation to merge with the present citation.
        :type acquisition: Citation
        """
        self._merge_privacy(acquisition)
        self._merge_note_list(acquisition)
        self._merge_media_list(acquisition)
        self._merge_tag_list(acquisition)
        # merge confidence
        level_priority = [0, 4, 1, 3, 2]
        idx = min(
            level_priority.index(self.confidence),
            level_priority.index(acquisition.confidence),
        )
        self.confidence = level_priority[idx]
        self._merge_attribute_list(acquisition)
        # N.B. a Citation can refer to only one 'Source', so the
        # 'Source' from acquisition cannot be merged in

    def set_confidence_level(self, val):
        """Set the confidence level."""
        self.confidence = val

    def get_confidence_level(self):
        """Return the confidence level."""
        return self.confidence

    def set_page(self, page):
        """Set the page indicator of the Citation."""
        self.page = page

    def get_page(self):
        """Get the page indicator of the Citation."""
        return self.page

    def set_reference_handle(self, val):
        """Set the source handle."""
        self.source_handle = val

    def get_reference_handle(self):
        """Get the source handle."""
        return self.source_handle



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import unittest
from ..treebasemodel import Node, NodeMap


class NodeTest(unittest.TestCase):
    def test_addremovechildren(self):
        n = Node("1", "", "key_to_sort_on", None, None)
        nm = NodeMap()
        nm.add_node(n)

        n2 = Node("2", "", "key_to_sort_on2", None, None)
        n.add_child(n2, nm)
        nm.add_node(n2)

        n3 = Node("2", "", "", None, None)
        n.add_child(n3, nm)
        nm.add_node(n3)

        n.remove_child(n3, nm)
        nm.del_node(n3)
        n.remove_child(n2, nm)
        nm.del_node(n2)
        self.assertEqual(len(n.children), 0)


if __name__ == "__main__":
    unittest.main()



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2013 Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Gtk modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.plug import Gramplet
from gramps.gui.widgets.styledtexteditor import StyledTextEditor
from gramps.gui.widgets import SimpleButton
from gramps.gui.dialog import WarningDialog
from gramps.gen.lib import StyledText, Note, NoteType
from gramps.gen.db import DbTxn
from gramps.gen.errors import WindowActiveError
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


class ToDo(Gramplet):
    """
    Displays the To Do notes for an object.
    """

    def init(self):
        self.gui.WIDGET = self.build_gui()
        self.gui.get_container_widget().remove(self.gui.textview)
        self.gui.get_container_widget().add(self.gui.WIDGET)
        self.gui.WIDGET.show()

    def build_gui(self):
        """
        Build the GUI interface.
        """
        top = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        hbox = Gtk.Box()
        self.left = SimpleButton("go-previous", self.left_clicked)
        self.left.set_tooltip_text(_("Previous To Do note"))
        self.left.set_sensitive(False)
        hbox.pack_start(self.left, False, False, 0)
        self.right = SimpleButton("go-next", self.right_clicked)
        self.right.set_tooltip_text(_("Next To Do note"))
        self.right.set_sensitive(False)
        hbox.pack_start(self.right, False, False, 0)
        self.edit = SimpleButton("gtk-edit", self.edit_clicked)
        self.edit.set_tooltip_text(_("Edit the selected To Do note"))
        self.edit.set_sensitive(False)
        hbox.pack_start(self.edit, False, False, 0)
        self.new = SimpleButton("document-new", self.new_clicked)
        self.new.set_tooltip_text(_("Add a new To Do note"))
        hbox.pack_start(self.new, False, False, 0)
        self.page = Gtk.Label()
        hbox.pack_end(self.page, False, False, 10)

        scrolledwindow = Gtk.ScrolledWindow()
        scrolledwindow.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self.texteditor = StyledTextEditor()
        self.texteditor.set_editable(False)
        self.texteditor.set_wrap_mode(Gtk.WrapMode.WORD)
        scrolledwindow.add(self.texteditor)

        top.pack_start(hbox, False, False, 0)
        top.pack_start(scrolledwindow, True, True, 0)
        top.show_all()
        return top

    def get_note_list(self, obj):
        """
        Get a list of To Do notes for the current object.
        """
        note_list = []
        for note_handle in obj.get_note_list():
            note = self.dbstate.db.get_note_from_handle(note_handle)
            if int(note.get_type()) == NoteType.TODO:
                note_list.append(note.get_handle())
        return note_list

    def get_notes(self, obj):
        """
        Display the To Do notes for the current object.
        """
        self.obj = obj
        self.left.set_sensitive(False)
        self.right.set_sensitive(False)
        self.edit.set_sensitive(False)
        self.texteditor.set_text(StyledText())
        self.note_list = self.get_note_list(obj)
        self.page.set_text("")
        if len(self.note_list) > 0:
            self.set_has_data(True)
            self.edit.set_sensitive(True)
            if len(self.note_list) > 1:
                self.right.set_sensitive(True)
            self.current = 0
            self.display_note()
        else:
            self.set_has_data(False)

    def clear_text(self):
        self.left.set_sensitive(False)
        self.right.set_sensitive(False)
        self.edit.set_sensitive(False)
        self.texteditor.set_text(StyledText())
        self.page.set_text("")
        self.current = 0

    def display_note(self):
        """
        Display the current note.
        """
        note_handle = self.note_list[self.current]
        note = self.dbstate.db.get_note_from_handle(note_handle)
        self.texteditor.set_text(note.get_styledtext())
        self.page.set_text(
            _("%(current)d of %(total)d")
            % {"current": self.current + 1, "total": len(self.note_list)}
        )

    def left_clicked(self, button):
        """
        Display the previous note.
        """
        if self.current > 0:
            self.current -= 1
            self.right.set_sensitive(True)
            if self.current == 0:
                self.left.set_sensitive(False)
            self.display_note()

    def right_clicked(self, button):
        """
        Display the next note.
        """
        if self.current < len(self.note_list) - 1:
            self.current += 1
            self.left.set_sensitive(True)
            if self.current == len(self.note_list) - 1:
                self.right.set_sensitive(False)
            self.display_note()

    def get_has_data(self, obj):
        """
        Return True if the gramplet has data, else return False.
        """
        if obj is None:
            return False
        if self.get_note_list(obj):
            return True
        return False

    def edit_clicked(self, obj):
        """
        Edit current To Do note.
        """
        from gramps.gui.editors import EditNote

        note_handle = self.note_list[self.current]
        note = self.dbstate.db.get_note_from_handle(note_handle)
        try:
            EditNote(self.gui.dbstate, self.gui.uistate, [], note)
        except WindowActiveError:
            pass

    def new_clicked(self, obj):
        """
        Create a new To Do note.
        """
        nav_type = self.uistate.viewmanager.active_page.navigation_type()
        active_handle = self.get_active(nav_type)
        if active_handle:
            from gramps.gui.editors import EditNote

            note = Note()
            note.set_type(NoteType.TODO)
            try:
                EditNote(self.gui.dbstate, self.gui.uistate, [], note, self.created)
            except WindowActiveError:
                pass
        else:
            WarningDialog(
                _("No active object"),
                _("First select the object to which you want to attach a note")
                + _(":")
                + _(nav_type),
                parent=self.uistate.window,
            )


class PersonToDo(ToDo):
    """
    Displays the To Do notes for a person.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "person-update", self.update)

    def active_changed(self, handle):
        self.update()

    def update_has_data(self):
        active_handle = self.get_active("Person")
        if active_handle:
            active = self.dbstate.db.get_person_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.clear_text()
        active_handle = self.get_active("Person")
        if active_handle:
            active = self.dbstate.db.get_person_from_handle(active_handle)
            if active:
                self.get_notes(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)

    def created(self, handle):
        with DbTxn("Attach Note", self.dbstate.db) as trans:
            self.obj.add_note(handle)
            self.dbstate.db.commit_person(self.obj, trans)


class EventToDo(ToDo):
    """
    Displays the To Do notes for an event.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "event-update", self.update)
        self.connect_signal("Event", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Event")
        if active_handle:
            active = self.dbstate.db.get_event_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.clear_text()
        active_handle = self.get_active("Event")
        if active_handle:
            active = self.dbstate.db.get_event_from_handle(active_handle)
            if active:
                self.get_notes(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)

    def created(self, handle):
        with DbTxn("Attach Note", self.dbstate.db) as trans:
            self.obj.add_note(handle)
            self.dbstate.db.commit_event(self.obj, trans)


class FamilyToDo(ToDo):
    """
    Displays the To Do notes for a family.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "family-update", self.update)
        self.connect_signal("Family", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Family")
        if active_handle:
            active = self.dbstate.db.get_family_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.clear_text()
        active_handle = self.get_active("Family")
        if active_handle:
            active = self.dbstate.db.get_family_from_handle(active_handle)
            if active:
                self.get_notes(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)

    def created(self, handle):
        with DbTxn("Attach Note", self.dbstate.db) as trans:
            self.obj.add_note(handle)
            self.dbstate.db.commit_family(self.obj, trans)


class PlaceToDo(ToDo):
    """
    Displays the To Do notes for a place.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "place-update", self.update)
        self.connect_signal("Place", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Place")
        if active_handle:
            active = self.dbstate.db.get_place_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.clear_text()
        active_handle = self.get_active("Place")
        if active_handle:
            active = self.dbstate.db.get_place_from_handle(active_handle)
            if active:
                self.get_notes(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)

    def created(self, handle):
        with DbTxn("Attach Note", self.dbstate.db) as trans:
            self.obj.add_note(handle)
            self.dbstate.db.commit_place(self.obj, trans)


class SourceToDo(ToDo):
    """
    Displays the To Do notes for a source.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "source-update", self.update)
        self.connect_signal("Source", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Source")
        if active_handle:
            active = self.dbstate.db.get_source_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.clear_text()
        active_handle = self.get_active("Source")
        if active_handle:
            active = self.dbstate.db.get_source_from_handle(active_handle)
            if active:
                self.get_notes(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)

    def created(self, handle):
        with DbTxn("Attach Note", self.dbstate.db) as trans:
            self.obj.add_note(handle)
            self.dbstate.db.commit_source(self.obj, trans)


class CitationToDo(ToDo):
    """
    Displays the To Do notes for a Citation.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "citation-update", self.update)
        self.connect_signal("Citation", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Citation")
        if active_handle:
            active = self.dbstate.db.get_citation_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.clear_text()
        active_handle = self.get_active("Citation")
        if active_handle:
            active = self.dbstate.db.get_citation_from_handle(active_handle)
            if active:
                self.get_notes(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)

    def created(self, handle):
        with DbTxn("Attach Note", self.dbstate.db) as trans:
            self.obj.add_note(handle)
            self.dbstate.db.commit_citation(self.obj, trans)


class RepositoryToDo(ToDo):
    """
    Displays the To Do notes for a repository.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "repository-update", self.update)
        self.connect_signal("Repository", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Repository")
        if active_handle:
            active = self.dbstate.db.get_repository_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.clear_text()
        active_handle = self.get_active("Repository")
        if active_handle:
            active = self.dbstate.db.get_repository_from_handle(active_handle)
            if active:
                self.get_notes(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)

    def created(self, handle):
        with DbTxn("Attach Note", self.dbstate.db) as trans:
            self.obj.add_note(handle)
            self.dbstate.db.commit_repository(self.obj, trans)


class MediaToDo(ToDo):
    """
    Displays the To Do notes for a media object.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "media-update", self.update)
        self.connect_signal("Media", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Media")
        if active_handle:
            active = self.dbstate.db.get_media_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.clear_text()
        active_handle = self.get_active("Media")
        if active_handle:
            active = self.dbstate.db.get_media_from_handle(active_handle)
            if active:
                self.get_notes(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)

    def created(self, handle):
        with DbTxn("Attach Note", self.dbstate.db) as trans:
            self.obj.add_note(handle)
            self.dbstate.db.commit_media(self.obj, trans)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ._isdescendantof import IsDescendantOf
from ._matchesfilter import MatchesFilter


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from typing import Set
from ....lib import Person
from ....db import Database


# -------------------------------------------------------------------------
#
# IsDescendantOfFilterMatch
#
# -------------------------------------------------------------------------
class IsDescendantOfFilterMatch(IsDescendantOf):
    """Rule that checks for a person that is a descendant
    of someone matched by a filter"""

    labels = [_("Filter name:")]
    name = _("Descendants of <filter> match")
    category = _("Descendant filters")
    description = _(
        "Matches people that are descendants " "of anybody matched by a filter"
    )

    def prepare(self, db: Database, user):
        self.db = db
        self.selected_handles: Set[str] = set()
        try:
            if int(self.list[1]):
                first = False
            else:
                first = True
        except IndexError:
            first = True

        self.filt = MatchesFilter(self.list[0:1])
        self.filt.requestprepare(db, user)
        if user:
            user.begin_progress(
                self.category,
                _("Retrieving all sub-filter matches"),
                db.get_number_of_people(),
            )
        for person in db.iter_people():
            if user:
                user.step_progress()
            if self.filt.apply_to_one(db, person):
                self.init_list(person, first)
        if user:
            user.end_progress()

    def reset(self):
        self.filt.requestreset()
        self.selected_handles.clear()

    def apply_to_one(self, db: Database, person: Person) -> bool:
        return person.handle in self.selected_handles



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2013       Doug Blank <doug.blank@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Researcher information for Gramps.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .locationbase import LocationBase


# -------------------------------------------------------------------------
#
# Researcher
#
# -------------------------------------------------------------------------
class Researcher(LocationBase):
    """
    Contains the information about the owner of the database.
    """

    def __init__(self, source=None):
        """
        Initialize the Researcher object, copying from the source if provided.
        """

        LocationBase.__init__(self, source)
        if source:
            self.name = source.name
            self.addr = source.addr
            self.email = source.email
        else:
            self.name = ""
            self.addr = ""
            self.email = ""

    def serialize(self):
        """
        Convert the object to a serialized tuple of data.
        """
        return (LocationBase.serialize(self), self.name, self.addr, self.email)

    def unserialize(self, data):
        """
        Convert a serialized tuple of data to an object.
        """
        (location, self.name, self.addr, self.email) = data
        LocationBase.unserialize(self, location)

        return self

    def set_name(self, data):
        """Set the database owner's name."""
        self.name = data

    def get_name(self):
        """Return the database owner's name."""
        return self.name

    def set_address(self, data):
        """Set the database owner's address."""
        self.addr = data

    def get_address(self):
        """Return the database owner's address."""
        return self.addr

    def set_email(self, data):
        """Set the database owner's email."""
        self.email = data

    def get_email(self):
        """Return the database owner's email."""
        return self.email

    def set_from(self, other_researcher):
        """Set all attributes from another instance."""
        self.street = other_researcher.street
        self.locality = other_researcher.locality
        self.city = other_researcher.city
        self.county = other_researcher.county
        self.state = other_researcher.state
        self.country = other_researcher.country
        self.postal = other_researcher.postal
        self.phone = other_researcher.phone

        self.name = other_researcher.name
        self.addr = other_researcher.addr
        self.email = other_researcher.email

    def get(self):
        """
        Return attributes about the researcher.
        """
        return [
            getattr(self, value)
            for value in [
                "name",
                "addr",
                "locality",
                "city",
                "state",
                "country",
                "postal",
                "phone",
                "email",
            ]
        ]

    def is_empty(self):
        """
        Check and return true if object empty.
        """
        for attr in [
            "name",
            "addr",
            "locality",
            "city",
            "state",
            "country",
            "postal",
            "phone",
            "email",
        ]:
            if getattr(self, attr) != "":
                return False
        return True

    def get_object_state(self):
        """
        Get the current object state as a dictionary.

        By default this returns the public attributes of the instance.  This
        method can be overridden if the class requires other attributes or
        properties to be saved.

        This method is called to provide the information required to serialize
        the object.

        :returns: Returns a dictionary of attributes that represent the state
                  of the object.
        :rtype: dict
        """
        attr_dict = dict(
            (key, value)
            for key, value in self.__dict__.items()
            if not key.startswith("_")
        )
        attr_dict["_class"] = self.__class__.__name__
        return attr_dict

    def set_object_state(self, attr_dict):
        """
        Set the current object state using information provided in the given
        dictionary.

        By default this sets the state of the object assuming that all items in
        the dictionary map to public attributes.  This method can be overridden
        to set the state using custom functionality.  For performance reasons
        it is useful to set a property without calling its setter function.  As
        JSON provides no distinction between tuples and lists, this method can
        also be use to convert lists into tuples where required.

        :param attr_dict: A dictionary of attributes that represent the state of
                          the object.
        :type attr_dict: dict
        """
        self.__dict__.update(attr_dict)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Filter rule to match persons with a particular event.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale
from ....lib.eventroletype import EventRoleType
from .._haseventbase import HasEventBase

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ....lib import Person
from ....db import Database


# -------------------------------------------------------------------------
#
# HasEvent
#
# -------------------------------------------------------------------------
class HasEvent(HasEventBase):
    """
    Rule that checks for a person with a event with a particular value.
    """

    labels = [
        _("Personal event:"),
        _("Date:"),
        _("Place:"),
        _("Description:"),
        _("Main Participants:"),
        _("Primary Role:"),
    ]
    name = _("People with the personal <event>")
    description = _("Matches people with a personal event of a particular value")

    def apply_to_one(self, db: Database, person: Person) -> bool:  # type: ignore[override]
        """
        Apply the rule. Return True if a match.
        """
        for event_ref in person.event_ref_list:
            if int(self.list[5]) and event_ref.role.value != EventRoleType.PRIMARY:
                # Only match primaries, no witnesses
                continue
            event = db.get_event_from_handle(event_ref.ref)
            if HasEventBase.apply_to_one(self, db, event):
                return True
        return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2025       David Straub
# Copyright (C) 2025       Steve Youngs
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
"""Types for static type checking."""

from __future__ import annotations

from typing import NewType, Type, TypeVar, Union

from .lib import (
    Citation,
    Event,
    Family,
    Media,
    Note,
    Person,
    Place,
    PrimaryObject,
    Repository,
    Source,
    Tag,
)
from .lib.tableobj import TableObject

PersonHandle = NewType("PersonHandle", str)
FamilyHandle = NewType("FamilyHandle", str)
EventHandle = NewType("EventHandle", str)
PlaceHandle = NewType("PlaceHandle", str)
SourceHandle = NewType("SourceHandle", str)
RepositoryHandle = NewType("RepositoryHandle", str)
CitationHandle = NewType("CitationHandle", str)
MediaHandle = NewType("MediaHandle", str)
NoteHandle = NewType("NoteHandle", str)
TagHandle = NewType("TagHandle", str)
PrimaryObjectHandle = Union[
    PersonHandle,
    FamilyHandle,
    EventHandle,
    PlaceHandle,
    SourceHandle,
    RepositoryHandle,
    CitationHandle,
    MediaHandle,
    NoteHandle,
]
AnyHandle = Union[PrimaryObjectHandle, TagHandle]
TableObjectType = TypeVar("TableObjectType", bound=TableObject)

PersonGrampsID = NewType("PersonGrampsID", str)
FamilyGrampsID = NewType("FamilyGrampsID", str)
EventGrampsID = NewType("EventGrampsID", str)
PlaceGrampsID = NewType("PlaceGrampsID", str)
SourceGrampsID = NewType("SourceGrampsID", str)
RepositoryGrampsID = NewType("RepositoryGrampsID", str)
CitationGrampsID = NewType("CitationGrampsID", str)
MediaGrampsID = NewType("MediaGrampsID", str)
NoteGrampsID = NewType("NoteGrampsID", str)
# No Tag IDs
PrimaryObjectGrampsID = Union[
    PersonGrampsID,
    FamilyGrampsID,
    EventGrampsID,
    PlaceGrampsID,
    SourceGrampsID,
    RepositoryGrampsID,
    CitationGrampsID,
    MediaGrampsID,
    NoteGrampsID,
]
AnyGrampsID = PrimaryObjectGrampsID  # No Tag IDs
